# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)



import os
import sys
import time
import glob
from tempfile import mkstemp
from shutil import move, rmtree

import numpy as np
import cv2
from natsort import natsorted
from confapp import conf


from idtrackerai.postprocessing.assign_them_all import close_trajectories_gaps

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.video")

class Video(object):
    """
    Attributes
    ----------

    video_path : str
        path to the video (if the video is split in portions, the path to one
        of the portions)
    number_of_animals : int
        number of animals to track
    episodes_start_end : list
        list of lists of the form [start, end] to parallelise processes
    original_bkg : ndarray
        model of the background with size the frame size
    bkg : ndarray
        resized bkg according to :attr:`.Video.resolution_reduction`
    subtract_bkg : bool
        True if bkg subtraction is performed
    original_ROI : ndarray
        ROI of the size of frame
    ROI : ndarray
        resized ROI according to :attr:`.Video.resolution_reduction`
    apply_ROI : bool
        True if a non-trivial ROI has been specified
    min_threshold : int
        minimum intensity threshold
    max_threshold : int
        maximum intensity threshold
    min_area : int
        minimum blob area
    max_area : int
        maximum blob area
    resegmentation_parameters : list
        segmentation parameters specific to a frame
    has_preprocessing_parameters : bool
        True if the preprocessing has been concluded succesfully
    maximum_number_of_blobs : int
        Maximum number of blobs segmented from a single frame
    resolution_reduction : float
        resolution reduction to be applied to the original frame
    erosion_kernel_size : int
        size of the kernel used to erode blobs while solving the crossings. See
        :mod:`~assign_them_all`
    """

    def __init__(self, video_path = None, open_multiple_files=False):
        logger.debug("Video object init")
        self._open_multiple_files = open_multiple_files
        self._video_path = video_path #string: path to the video
        self._number_of_animals = None #int: number of animals in the video
        self._episodes_start_end = None #list of lists: starting and ending frame per chunk [video is split for parallel computation]
        self._original_bkg = None #matrix [shape = shape of a frame] background used to do bkg subtraction
        self._bkg = None
        self._resolution_reduction = 1.
        self._subtract_bkg = None #boolean: True if the user specifies to subtract the background
        self._original_ROI = None #matrix [shape = shape of a frame] 255 are valid (part of the ROI) pixels and 0 are invalid according to openCV convention
        self._ROI = None
        self._apply_ROI = None #boolean: True if the user applies a ROI to the video
        self._min_threshold = conf.MIN_THRESHOLD_DEFAULT
        self._max_threshold = conf.MAX_THRESHOLD_DEFAULT
        self._min_area = conf.MIN_AREA_DEFAULT
        self._max_area = conf.MAX_AREA_DEFAULT
        self._resize = 1
        self._resegmentation_parameters = []
        self._tracking_interval = None
        self._has_preprocessing_parameters = False #boolean: True once the preprocessing parameters (max/min area, max/min threshold) are set and saved
        self._maximum_number_of_blobs = 0 #int: the maximum number of blobs detected in the video
        self._blobs_path = None #string: path to the saved list of blob objects
        self._blobs_path_segmented = None
        self._blobs_path_interpolated = None
        self._has_been_segmented = None
        self._has_been_preprocessed = None #boolean: True if a video has been fragmented in a past session
        self._preprocessing_folder = None
        self._images_folder = None
        self._global_fragments_path = None #string: path to saved list of global fragments
        self._has_been_pretrained = None
        self._pretraining_folder = None
        self._knowledge_transfer_model_folder = None
        self._identity_transfer = None
        self._tracking_with_knowledge_transfer = False
        self._percentage_of_accumulated_images = None
        self._first_accumulation_finished = None
        self._second_accumulation_finished = None
        self._has_been_assigned = None # ?
        self._has_duplications_solved = None
        self._has_crossings_solved = None
        self._has_trajectories = None
        self._has_trajectories_wo_gaps = None
        self._embeddings_folder = None # If embeddings are computed, the will be saved in this path
        self._first_frame_first_global_fragment = []
        self._segmentation_time = 0.
        self._crossing_detector_time = 0.
        self._fragmentation_time = 0.
        self._protocol1_time = 0.
        self._protocol2_time = 0.
        self._protocol3_pretraining_time = 0.
        self._protocol3_accumulation_time = 0.
        self._identify_time = 0.
        self._create_trajectories_time = 0.
        self._there_are_crossings = True
        self._track_wo_identities = False # Track without identities
        self._number_of_channels = 1
        if conf.SIGMA_GAUSSIAN_BLURRING is not None:
            self.sigma_gaussian_blurring = conf.SIGMA_GAUSSIAN_BLURRING

        # Flag to decide which type of interpolation is done. This flag is updated when we update a blob centroid
        self._is_centroid_updated = False

    @property
    def is_centroid_updated(self):
        return self._is_centroid_updated

    @is_centroid_updated.setter
    def is_centroid_updated(self, value):
        self._is_centroid_updated = value

    @property
    def number_of_channels(self):
        return self._number_of_channels

    @property
    def episodes_start_end(self):
        return self._episodes_start_end

    @property
    def has_preprocessing_parameters(self):
        return self._has_preprocessing_parameters

    @property
    def preprocessing_folder(self):
        return self._preprocessing_folder

    @property
    def images_folder(self):
        return self._images_folder

    @property
    def has_been_preprocessed(self):
        return self._has_been_preprocessed

    @property
    def has_been_segmented(self):
        return self._has_been_segmented

    @property
    def crossings_detector_folder(self):
        return self._crossings_detector_folder

    @property
    def has_been_pretrained(self):
        return self._has_been_pretrained

    @property
    def previous_session_folder(self):
        return self._previous_session_folder

    @property
    def pretraining_folder(self):
        return self._pretraining_folder

    @property
    def first_accumulation_finished(self):
        return self._first_accumulation_finished

    @property
    def second_accumulation_finished(self):
        return self._second_accumulation_finished

    @property
    def accumulation_folder(self):
        return self._accumulation_folder

    @property
    def percentage_of_accumulated_images(self):
        return self._percentage_of_accumulated_images

    @property
    def has_been_assigned(self):
        return self._has_been_assigned

    @property
    def has_duplications_solved(self):
        return self._has_duplications_solved

    @property
    def has_crossings_solved(self):
        return self._has_crossings_solved

    @property
    def has_trajectories(self):
        return self._has_trajectories

    @property
    def has_trajectories_wo_gaps(self):
        return self._has_trajectories_wo_gaps

    @property
    def embeddings_folder(self):
        return self._embeddings_folder

    @property
    def session_folder(self):
        return self._session_folder

    @property
    def apply_ROI(self):
        return self._apply_ROI

    @property
    def original_ROI(self):
        return self._original_ROI

    @property
    def ROI(self):
        return self._ROI

    @property
    def subtract_bkg(self):
        return self._subtract_bkg

    @property
    def original_bkg(self):
        return self._original_bkg

    @property
    def bkg(self):
        return self._bkg

    @property
    def resolution_reduction(self):
        return self._resolution_reduction

    @resolution_reduction.setter
    def resolution_reduction(self, value):
        self._resolution_reduction = value
        fake_frame = np.ones((self._original_height, self._original_width)).astype('uint8')
        fake_frame = cv2.resize(fake_frame, None,
                                fx = value,
                                fy = value,
                                interpolation = cv2.INTER_AREA)
        self._height = fake_frame.shape[0]
        self._width = fake_frame.shape[1]
        if self.subtract_bkg:
            self._bkg = cv2.resize(self.original_bkg, None,
                                            fx = value,
                                            fy = value,
                                            interpolation = cv2.INTER_AREA)
        if self.apply_ROI or self.original_ROI is not None:
            self._ROI = cv2.resize(self.original_ROI, None,
                                            fx = value,
                                            fy = value,
                                            interpolation = cv2.INTER_AREA)

    @property
    def min_threshold(self):
        return self._min_threshold

    @property
    def max_threshold(self):
        return self._max_threshold

    @property
    def min_area(self):
        return self._min_area

    @property
    def max_area(self):
        return self._max_area

    @property
    def resegmentation_parameters(self):
        return self._resegmentation_parameters

    @property
    def tracking_interval(self):
        """ Tuple with the starting and ending frame on which the tracking will
        be performed.
        """
        return self._tracking_interval

    @property
    def resize(self):
        return self._resize

    @property
    def blobs_path(self):
        """get the path to save the blob collection after segmentation.
        It checks that the segmentation has been succesfully performed"""
        if self.preprocessing_folder is None: return None
        self._blobs_path = os.path.join(self.preprocessing_folder, 'blobs_collection.npy')
        return self._blobs_path

    @property
    def blobs_path_segmented(self):
        """get the path to save the blob collection after segmentation.
        It checks that the segmentation has been succesfully performed"""
        self._blobs_path_segmented = os.path.join(self.preprocessing_folder, 'blobs_collection_segmented.npy')
        return self._blobs_path_segmented

    @property
    def blobs_path_interpolated(self):
        self._blobs_path_interpolated = os.path.join(self.preprocessing_folder, 'blobs_collection_interpolated.npy')
        return self._blobs_path_interpolated

    @property
    def global_fragments_path(self):
        """get the path to save the list of global fragments after
        fragmentation"""
        self._global_fragments_path = os.path.join(self.preprocessing_folder, 'global_fragments.npy')
        return self._global_fragments_path

    @property
    def fragments_path(self):
        """get the path to save the list of global fragments after
        fragmentation"""
        self._fragments_path = os.path.join(self.preprocessing_folder, 'fragments.npy')
        return self._fragments_path

    @property
    def erosion_kernel_size(self):
        return self._erosion_kernel_size

    @property
    def path_to_video_object(self):
        return self._path_to_video_object

    @property
    def git_commit(self):
        return '0'
        return self._git_commit

    def save(self):
        """save class"""
        logger.info("saving video object in %s" %self.path_to_video_object)
        np.save(self.path_to_video_object, self)

    @property
    def knowledge_transfer_model_folder(self):
        return self._knowledge_transfer_model_folder

    @knowledge_transfer_model_folder.setter
    def knowledge_transfer_model_folder(self, new_kt_model_path):
        if new_kt_model_path:
            subfolders = glob.glob(os.path.join(new_kt_model_path, "*"))
            if os.path.join(new_kt_model_path, "conv") in subfolders and os.path.join(new_kt_model_path, "softmax") in subfolders:
                self._knowledge_transfer_model_folder = new_kt_model_path
            else:
                raise ValueError("The model folders " + os.path.join(new_kt_model_path, "conv") + " and " + os.path.join(new_kt_model_path, "softmax") + " are missing")
        else:
            self._knowledge_transfer_model_folder = None

    @property
    def identity_transfer(self):
        return self._identity_transfer

    def is_identity_transfer_possible(self):
        return self.number_of_animals == self.knowledge_transfer_info_dict['number_of_animals']

    def check_and_set_identity_transfer_if_possible(self):
        if conf.KNOWLEDGE_TRANSFER_FOLDER_IDCNN is None:
            raise ValueError("Need to provide a value (path) for the varialbe KNOWLEDGE_TRANSFER_FOLDER_IDCNN in the local_settings.py file")
        self.knowledge_transfer_model_folder = conf.KNOWLEDGE_TRANSFER_FOLDER_IDCNN
        self.knowledge_transfer_info_dict = np.load(os.path.join(self.knowledge_transfer_model_folder, 'info.npy'), allow_pickle=True).item()
        if self.is_identity_transfer_possible():
            logger.info("Tracking with identity transfer. The IDENTIFICATION_IMAGE_SIZE will be matched\
                        to the input_image_size of the transferred network")
            self._identity_transfer = True
            self._tracking_with_knowledge_transfer = True
            self._knowledge_transfer_model_folder = conf.KNOWLEDGE_TRANSFER_FOLDER_IDCNN
            conf.IDENTIFICATION_IMAGE_SIZE = self.knowledge_transfer_info_dict['input_image_size']
        else:
            logger.info("Tracking with identity transfer failed. The number of animals in the video\
                        needs to be the same as the number of animals in the transferred network")
            self._identity_transfer = False
            self._tracking_with_knowledge_transfer = False
            conf.KNOWLEDGE_TRANSFER_FOLDER_IDCNN = None

    @property
    def tracking_with_knowledge_transfer(self):
        return self._tracking_with_knowledge_transfer

    @property
    def track_wo_identities(self):
        """
        Flag track without identities

        returns: Boolean
        """
        return self._track_wo_identities

    @property
    def video_path(self):
        return self._video_path

    @video_path.setter
    def video_path(self, value):
        video_name, video_extension = os.path.splitext(value)
        if video_extension in conf.AVAILABLE_VIDEO_EXTENSION:
            self._video_path = value
            #get video folder
            self._video_folder = os.path.dirname(self.video_path)
            #collect some info on the video (resolution, number of frames, ..)
            if not hasattr(self,'number_of_frames'):
                self.get_info()
            self.modified = time.strftime("%c")
        else:
            raise ValueError("Supported video extensions are ", conf.AVAILABLE_VIDEO_EXTENSION)

    @property
    def video_folder(self):
        return self._video_folder

    @property
    def number_of_animals(self):
        return self._number_of_animals

    @property
    def maximum_number_of_blobs(self):
        return self._maximum_number_of_blobs

    @property
    def number_of_frames(self):
        return self._number_of_frames

    @property
    def number_of_episodes(self):
        return self._number_of_episodes

    @property
    def open_multiple_files(self):
        return self._open_multiple_files

    def check_split_video(self):
        """If the video is divided in segments retrieves their paths
        """
        paths_to_video_segments = scanFolder(self.video_path)
        if len(paths_to_video_segments) > 1 and self.open_multiple_files:
            return paths_to_video_segments
        else:
            return None

    @property
    def paths_to_video_segments(self):
        return self._paths_to_video_segments

    @property
    def original_width(self):
        return self._original_width

    @property
    def original_height(self):
        return self._original_height

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def frames_per_second(self):
        return self._frames_per_second

    @property
    def fragment_identifier_to_index(self):
        return self._fragment_identifier_to_index

    @property
    def first_frame_first_global_fragment(self):
        return self._first_frame_first_global_fragment

    @property
    def median_body_length(self):
        return self._median_body_length

    @property
    def median_body_length_full_resolution(self):
        return self.median_body_length/self.resolution_reduction

    @property
    def model_area(self):
        return self._model_area

    @property
    def maximum_number_of_images_in_global_fragments(self):
        return self._maximum_number_of_images_in_global_fragments

    @property
    def number_of_unique_images_in_global_fragments(self):
        return self._number_of_unique_images_in_global_fragments

    @property
    def there_are_crossings(self):
        return self._there_are_crossings

    @property
    def ratio_accumulated_images(self):
        return self._ratio_accumulated_images

    @property
    def segmentation_time(self):
        return self._segmentation_time

    @property
    def crossing_detector_time(self):
        return self._crossing_detector_time

    @property
    def fragmentation_time(self):
        return self._fragmentation_time

    @property
    def protocol1_time(self):
        return self._protocol1_time

    @property
    def protocol2_time(self):
        return self._protocol2_time

    @property
    def protocol3_pretraining_time(self):
        return self._protocol3_pretraining_time

    @property
    def protocol3_accumulation_time(self):
        return self._protocol3_accumulation_time

    @property
    def identify_time(self):
        return self._identify_time

    @property
    def create_trajectories_time(self):
        return self._create_trajectories_time

    def check_paths_consistency_with_video_path(self, new_video_path):
        if self.video_path != new_video_path:
            self.update_paths(new_video_path)

    def update_paths(self, new_video_object_path):
        if new_video_object_path == '': raise ValueError("The path to the video object is an empty string")
        new_session_path = os.path.split(new_video_object_path)[0]
        old_session_path = self.session_folder
        video_name = os.path.split(self._video_path)[1]
        self._video_folder = os.path.split(new_session_path)[0]
        self.video_path = os.path.join(self.video_folder,video_name)
        attributes_to_modify = {key: getattr(self, key) for key in self.__dict__
        if isinstance(getattr(self, key), str)
        and old_session_path in getattr(self, key) }

        for key in attributes_to_modify:
            new_value = attributes_to_modify[key].replace(old_session_path, new_session_path)
            setattr(self, key, new_value)

        if self.paths_to_video_segments is not None and len(self.paths_to_video_segments) != 0:
            logger.info("Updating paths_to_video_segments")
            new_paths_to_video_segments = []
            for path in self.paths_to_video_segments:
                new_path = os.path.join(self.video_folder, os.path.split(path)[1])
                new_paths_to_video_segments.append(new_path)
            self._paths_to_video_segments = new_paths_to_video_segments

        logger.info("Updating checkpoint files")
        folders_to_check = ['crossings_detector_folder', 'pretraining_folder', 'accumulation_folder']
        for folder in folders_to_check:
            if hasattr(self, folder) and getattr(self, folder) is not None:
                if folder == folders_to_check[0]:
                    checkpoint_path = os.path.join(self.crossings_detector_folder, 'checkpoint')
                    if os.path.isfile(checkpoint_path):
                        self.update_tensorflow_checkpoints_file(checkpoint_path, old_session_path, new_session_path)
                    else:
                        logger.warn('No checkpoint found in %s ' %folder)
                else:
                    for sub_folder in ['conv', 'softmax']:
                        checkpoint_path = os.path.join(getattr(self, folder), sub_folder, 'checkpoint')
                        if os.path.isfile(checkpoint_path):
                            self.update_tensorflow_checkpoints_file(checkpoint_path, old_session_path, new_session_path)
                        else:
                            logger.warn('No checkpoint found in %s ' %os.path.join(getattr(self,folder),sub_folder))

        logger.info("Saving video object")
        self.save()
        logger.info("Done")

    @staticmethod
    def update_tensorflow_checkpoints_file(checkpoint_path, current_session_name, new_session_name):
        #checkpoint_file = open(checkpoint_path, "r")
        fh, abs_path = mkstemp()
        with os.fdopen(fh,'w') as new_file:
            with open(checkpoint_path) as old_file:
                for line in old_file:
                    splitted_line = line.split('"')
                    string_to_replace = splitted_line[1]
                    new_string = string_to_replace.replace(current_session_name, new_session_name)
                    splitted_line[1] = new_string
                    new_line = '"'.join(splitted_line)
                    new_file.write(new_line)

        os.remove(checkpoint_path)
        move(abs_path, checkpoint_path)

    def rename_session_folder(self, new_session_name):
        assert new_session_name != ''
        new_session_name = 'session_' + new_session_name
        current_session_name = os.path.split(self.session_folder)[1]
        logger.info("Updating checkpoint files")
        folders_to_check = ['video_folder',
                            'preprocessing_folder',
                            'logs_folder',
                            'previous_session_folder'
                            'crossings_detector_folder',
                            'pretraining_folder',
                            'accumulation_folder']

        for folder in folders_to_check:
            if hasattr(self, folder) and getattr(self, folder) is not None:
                if folder == folders_to_check[0]:
                    checkpoint_path = os.path.join(self.crossings_detector_folder, 'checkpoint')
                    if os.path.isfile(checkpoint_path):
                        self.update_tensorflow_checkpoints_file(checkpoint_path, current_session_name, new_session_name)
                    else:
                        logger.warn('No checkpoint found in %s ' %folder)
                else:
                    for sub_folder in ['conv', 'softmax']:
                        checkpoint_path = os.path.join(getattr(self,folder), sub_folder, 'checkpoint')
                        if os.path.isfile(checkpoint_path):
                            self.update_tensorflow_checkpoints_file(checkpoint_path, current_session_name, new_session_name)
                        else:
                            logger.warn('No checkpoint found in %s ' %os.path.join(getattr(self, folder), sub_folder))

        attributes_to_modify = {key: getattr(self, key) for key in self.__dict__
        if isinstance(getattr(self, key), str)
        and current_session_name in getattr(self, key) }
        logger.info("Modifying folder name from %s to %s "  %(current_session_name, new_session_name))
        os.rename(self.session_folder,
                os.path.join(self.video_folder, new_session_name))
        logger.info("Updating video object")

        for key in attributes_to_modify:
            new_value = attributes_to_modify[key].replace(current_session_name, new_session_name)
            setattr(self, key, new_value)

        logger.info("Saving video object")
        self.save()

    def get_info(self):
        """Retrieves basic information concerning the video. If the video is
        recorded as a single file, it generates "episodes" for parallelisation
        during segmentation
        """
        self._paths_to_video_segments = self.check_split_video()
        cap = cv2.VideoCapture(self.video_path)
        self._width  = self._original_width = int(cap.get(3))
        self._height = self._original_height = int(cap.get(4))

        try:
            self._frames_per_second = int(cap.get(5))
        except:
            self._frames_per_second = None
            logger.info("Cannot read frame per second")
        if self._paths_to_video_segments is None:
            self._number_of_frames = int(cap.get(7))
            self.get_episodes()
        else:
            chunks_lengths = [int(cv2.VideoCapture(chunk).get(7)) for chunk in self._paths_to_video_segments]
            self._episodes_start_end = [(np.sum(chunks_lengths[:i-1], dtype = np.int), np.sum(chunks_lengths[:i])-1) for i in range(1,len(chunks_lengths)+1)]
            self._number_of_frames = np.sum(chunks_lengths)
            self._number_of_episodes = len(self._paths_to_video_segments)
        cap.release()

    @property
    def identification_image_size(self):
        return self._identification_image_size

    def compute_identification_image_size(self, maximum_body_length):
        """Uses an estimate of the body length of the animals in order to
        compute the size of the square image that is generated from every
        blob to identify the animals
        """
        if conf.IDENTIFICATION_IMAGE_SIZE is None:
            identification_image_size = int(maximum_body_length / np.sqrt(2))
            identification_image_size = identification_image_size + identification_image_size % 2
            self._identification_image_size = (identification_image_size, identification_image_size, self.number_of_channels)
        else:
            self._identification_image_size = conf.IDENTIFICATION_IMAGE_SIZE

    def init_processes_time_attributes(self):
        self.generate_trajectories_time = 0
        self.solve_impossible_jumps_time = 0
        self.solve_duplications_time = 0
        self.assignment_time = 0
        self.second_accumulation_time = 0
        self.pretraining_time = 0
        self.assignment_time = 0
        self.first_accumulation_time = 0
        self.preprocessing_time = 0

    def create_session_folder(self, name = ''):
        """Creates a folder named training in video_folder and a folder session_num
        where num is the session number and it is created everytime one starts
        training a network for a certain video_path
        """
        if name == '':
            self._session_folder = os.path.join(self.video_folder, 'session')
        else:
            self._session_folder = os.path.join(self.video_folder, 'session_' + name)

        if not os.path.isdir(self._session_folder):
            os.makedirs(self._session_folder)
            self._previous_session_folder = ''
        else:
            self._previous_session_folder = self.session_folder

        self._path_to_video_object = os.path.join(self.session_folder, 'video_object.npy')
        logger.info("the folder %s has been created" %self.session_folder)

    def create_images_folders(self):
        """Create a folder named images inside of the session_folder
        """
        ## for RAM optimization
        self._segmentation_data_folder = os.path.join(self.session_folder, 'segmentation_data')
        self._identification_images_folder = os.path.join(self.session_folder, 'identification_images')
        self.identification_images_file_path = os.path.join(self._identification_images_folder, 'i_images.hdf5')
        if not os.path.isdir(self._segmentation_data_folder):
            os.makedirs(self._segmentation_data_folder)
            logger.info("the folder %s has been created" %self._segmentation_data_folder)

        if not os.path.isdir(self._identification_images_folder):
            os.makedirs(self._identification_images_folder)
            logger.info("the folder %s has been created" %self._identification_images_folder)


    def create_preprocessing_folder(self):
        """If it does not exist creates a folder called preprocessing
        in the video folder"""
        self._preprocessing_folder = os.path.join(self.session_folder, 'preprocessing')
        if not os.path.isdir(self.preprocessing_folder):
            os.makedirs(self.preprocessing_folder)
            logger.info("the folder %s has been created" %self._preprocessing_folder)


    def create_crossings_detector_folder(self):
        """If it does not exist creates a folder called crossing_detector
        in the video folder"""
        logger.info('setting path to save crossing detector model')
        self._crossings_detector_folder = os.path.join(self.session_folder, 'crossings_detector')
        if not os.path.isdir(self.crossings_detector_folder):
            logger.info("the folder %s has been created" %self.crossings_detector_folder)
            os.makedirs(self.crossings_detector_folder)

    def create_pretraining_folder(self, delete = False):
        """Creates a folder named pretraining in video_folder where the model
        trained during the pretraining is stored
        """
        self._pretraining_folder = os.path.join(self.session_folder, 'pretraining')
        if not os.path.isdir(self.pretraining_folder):
            os.makedirs(self.pretraining_folder)
        elif delete:
            rmtree(self.pretraining_folder)
            os.makedirs(self.pretraining_folder)


    def create_accumulation_folder(self, iteration_number = 0, delete = False):
        """Folder in which the model generated while accumulating is stored
        (after pretraining)
        """
        accumulation_folder_name = 'accumulation_' + str(iteration_number)
        self._accumulation_folder = os.path.join(self.session_folder, accumulation_folder_name)
        if not os.path.isdir(self.accumulation_folder):
            os.makedirs(self.accumulation_folder)
        elif delete:
            rmtree(self.accumulation_folder)
            os.makedirs(self.accumulation_folder)

    def init_accumulation_statistics_attributes(self, attributes = None):
        if attributes is None:
            attributes = ['number_of_accumulated_global_fragments',
                        'number_of_non_certain_global_fragments',
                        'number_of_randomly_assigned_global_fragments',
                        'number_of_nonconsistent_global_fragments',
                        'number_of_nonunique_global_fragments',
                        'number_of_acceptable_global_fragments',
                        'validation_accuracy','validation_individual_accuracies',
                        'training_accuracy','training_individual_accuracies',
                        'ratio_of_accumulated_images']
        self.accumulation_statistics_attributes_list = attributes
        [setattr(self, attribute, []) for attribute in self.accumulation_statistics_attributes_list]

    def store_accumulation_step_statistics_data(self, new_values):
        [getattr(self, attr).append(value) for attr, value in zip(self.accumulation_statistics_attributes_list, new_values)]

    def store_accumulation_statistics_data(self, accumulation_trial, number_of_possible_accumulation = conf.MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS + 1):
        if not hasattr(self, 'accumulation_statistics'): self.accumulation_statistics = [None] * number_of_possible_accumulation
        self.accumulation_statistics[accumulation_trial] = [getattr(self, stat_attr)
                                                            for stat_attr in self.accumulation_statistics_attributes_list]

    # @property
    # def final_training_folder(self):
    #     return self._final_training_folder
    #
    # def create_training_folder(self):
    #     """Folder in which the last model is stored (after accumulation)
    #     """
    #     self._final_training_folder = os.path.join(self.session_folder, 'training')
    #     if not os.path.isdir(self.final_training_folder):
    #         os.makedirs(self.final_training_folder)

    def create_trajectories_folder(self):
        """Folder in which trajectories files are stored
        """
        self.trajectories_folder = os.path.join(self.session_folder,'trajectories')
        if not os.path.isdir(self.trajectories_folder):
            logger.info("Creating trajectories folder...")
            os.makedirs(self.trajectories_folder)
            logger.info("the folder %s has been created" %self.trajectories_folder)

    def create_trajectories_wo_identities_folder(self):
        """Folder in which trajectories without identites are stored
        """
        self.trajectories_wo_identities_folder = os.path.join(self.session_folder,'trajectories_wo_identities')
        if not os.path.isdir(self.trajectories_wo_identities_folder):
            logger.info("Creating trajectories folder...")
            os.makedirs(self.trajectories_wo_identities_folder)
            logger.info("the folder %s has been created" %self.trajectories_wo_identities_folder)

    def create_trajectories_wo_gaps_folder(self):
        """Folder in which trajectories files are stored
        """
        self.trajectories_wo_gaps_folder = os.path.join(self.session_folder, 'trajectories_wo_gaps')
        if not os.path.isdir(self.trajectories_wo_gaps_folder):
            logger.info("Creating trajectories folder...")
            os.makedirs(self.trajectories_wo_gaps_folder)
            logger.info("the folder %s has been created" %self.trajectories_wo_gaps_folder)

    def create_embeddings_folder(self):
        """If it does not exist creates a folder called embedding
        in the video folder"""
        self._embeddings_folder = os.path.join(self.session_folder, 'embeddings')
        if not os.path.isdir(self.embeddings_folder):
            os.makedirs(self.embeddings_folder)
            logger.info("the folder %s has been created" %self.embeddings_folder)

    def get_episodes(self):
        """Split video in episodes (chunks) of FRAMES_PER_EPISODE frames
        for parallelisation"""
        starting_frames = np.arange(0, self.number_of_frames, conf.FRAMES_PER_EPISODE)
        ending_frames = np.hstack((starting_frames[1:]-1, self.number_of_frames))
        self._episodes_start_end = list(zip(starting_frames, ending_frames))
        self._number_of_episodes = len(starting_frames)

    def in_which_episode(self, frame_number):
        """Check to which episode a frame index belongs
        """
        episode_number = [i for i, episode_start_end in enumerate(self._episodes_start_end)
                            if episode_start_end[0] <= frame_number
                            and episode_start_end[1] >= frame_number]
        if episode_number:
            return episode_number[0]
        else:
            return None

    def copy_attributes_between_two_video_objects(self, video_object_source, list_of_attributes, is_property = None):
        for i, attribute in enumerate(list_of_attributes):
            if not hasattr(video_object_source, attribute):
                logger.warning("attribute %s does not exist" %attribute)
            else:
                if is_property is not None:
                    attribute_is_property = is_property[i]
                    if attribute_is_property:
                        setattr(self, '_' + attribute, getattr(video_object_source, attribute))
                    else:
                        setattr(self, attribute, getattr(video_object_source, attribute))
                else:
                    setattr(self, '_' + attribute, getattr(video_object_source, attribute))

    def compute_overall_P2(self, fragments):
        weighted_P2 = 0
        number_of_individual_blobs = 0

        for fragment in fragments:
            if fragment.is_an_individual:
                if fragment.assigned_identities[0] != 0:
                    weighted_P2 += fragment.P2_vector[fragment.assigned_identities[0] - 1] * fragment.number_of_images
                number_of_individual_blobs += fragment.number_of_images

        self.overall_P2 = weighted_P2 / number_of_individual_blobs

    def delete_accumulation_folders(self):
        for path in glob.glob(os.path.join(self.session_folder, '*')):
            if 'accumulation_' in path or 'pretraining' in path:
                rmtree(path, ignore_errors=True)

    def delete_data(self):
        logger.info("Data policy: {}".format(conf.DATA_POLICY))
        if conf.DATA_POLICY in ['trajectories', 'validation',
                           'knowledge_transfer', 'idmatcher.ai']:

            if os.path.isdir(self._segmentation_data_folder):
                logger.info("Deleting segmentation images")
                rmtree(self._segmentation_data_folder, ignore_errors=True)
            if os.path.isfile(self.global_fragments_path):
                logger.info("Deleting global fragments")
                os.remove(self.global_fragments_path)
            if os.path.isfile(self.blobs_path_segmented):
                logger.info("Deleting blobs segmented")
                os.remove(self.blobs_path_segmented)
            if hasattr(self, '_crossings_detector_folder') and os.path.isdir(self.crossings_detector_folder):
                logger.info("Deleting crossing detector folder")
                rmtree(self.crossings_detector_folder, ignore_errors=True)


        if conf.DATA_POLICY in ['trajectories', 'validation',
                           'knowledge_transfer']:
            if os.path.isdir(self._identification_images_folder):
                logger.info("Deleting identification images")
                rmtree(self._identification_images_folder, ignore_errors=True)

        if conf.DATA_POLICY in ['trajectories', 'validation']:
            logger.info("Deleting CNN models folders")
            self.delete_accumulation_folders()

        if conf.DATA_POLICY == 'trajectories':
            if os.path.isdir(self.preprocessing_folder):
                logger.info("Deleting preprocessing data")
                rmtree(self.preprocessing_folder, ignore_errors=True)


    def interpolate(self, list_of_blobs=None, list_of_fragments=None, identity=None, start=None, end=None):
        """
        Does the blobs positions interpolation

        :param ListOfBlobs list_of_blobs:
        :param identity:
        :param int start:
        :param int end:
        :param ListOfFragments list_of_fragments:
        """

        if self.is_centroid_updated:
            assert identity is not None and start is not None and end is not None and list_of_blobs is not None

            list_of_blobs.interpolate_from_user_generated_centroids(self, identity, start, end)

        else:
            assert list_of_fragments is not None and list_of_blobs is not None

            close_trajectories_gaps(self, list_of_blobs, list_of_fragments)

        self._is_centroid_updated = False


    def get_first_frame(self, list_of_blobs):
        if self.number_of_animals != 1:
            return self.first_frame_first_global_fragment[self.accumulation_trial]
        elif self.number_of_animals == 1:
            return 0
        else:
            for blobs_in_frame in list_of_blobs.blobs_in_video:
                if len(blobs_in_frame) != 0:
                    return blobs_in_frame[0].frame_number


def scanFolder(path):
    video = os.path.basename(path)
    filename, extension = os.path.splitext(video)
    folder = os.path.dirname(path)
    paths = glob.glob(folder + "/*" + extension)
    return natsorted(paths)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys
import numpy as np
from pprint import pprint
from idtrackerai.list_of_blobs import ListOfBlobs
from idtrackerai.blob import Blob
from idtrackerai.groundtruth_utils.generate_groundtruth import GroundTruth, GroundTruthBlob

"""Given two list of blobs_in_video, one deduced from human groundtruth and the other
generated by the tracking algorithm, compares them and gives back some statistics

Crossing: crossings are a special case. We ...
"""

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.compute_statistics_against_groundtruth")

def compare_tracking_against_groundtruth_no_gaps(number_of_animals,
                                                groundtruth, blobs_in_video_groundtruth,
                                                blobs_in_video, identities_dictionary_permutation):
    """
    blobs_in_video_groundtruth: cut groundtruth.blobs_in_video using start and end of the groundtruth object
    blobs_in_video: cut list_of_blobs.blobs_in_video using start and end of the groundtruth object
    """
    results = {}
    results['number_of_blobs_per_identity'] = {identity: 0 for identity in range(1, number_of_animals + 1)}
    results['number_of_individuals_badly_assigned'] = {identity: 0 for identity in range(1, number_of_animals + 1)}
    if identities_dictionary_permutation is not None:
        results['number_of_individuals_badly_interpolated'] = {
            identities_dictionary_permutation[identity]: groundtruth.wrong_crossing_counter[identity]
            for identity in groundtruth.wrong_crossing_counter}
        results['number_of_individuals_unidentified'] = {
            identities_dictionary_permutation[identity]: groundtruth.unidentified_individuals_counter[identity]
            for identity in groundtruth.wrong_crossing_counter}
    else:
        results['number_of_individuals_badly_interpolated'] = groundtruth.wrong_crossing_counter
        results['number_of_individuals_unidentified'] = groundtruth.unidentified_individuals_counter

    for blobs_in_frame_gt, blobs_in_frame in zip(blobs_in_video_groundtruth, blobs_in_video):
        for blob_gt, blob in zip(blobs_in_frame_gt, blobs_in_frame):

            if identities_dictionary_permutation is not None:
                if isinstance(blob_gt.identity, int):
                    gt_identity = identities_dictionary_permutation[blob_gt.identity]
                elif isinstance(blob_gt.identity, list):
                    gt_identity = [identities_dictionary_permutation[identity]
                                    for identity in blob_gt.identity]
            else:
                gt_identity = blob_gt.identity

            if isinstance(gt_identity, int) and gt_identity != 0:
                results['number_of_blobs_per_identity'][gt_identity] += 1
            elif isinstance(gt_identity, list):
                for identity in gt_identity:
                    if identity != 0:
                        results['number_of_blobs_per_identity'][identity] += 1
            elif gt_identity is None:
                logger.debug('***************************************unidentified blobs')

            if blob_gt.is_an_individual and not blob_gt.was_a_crossing:
                if gt_identity != blob.assigned_identities:
                    results['number_of_individuals_badly_assigned'][gt_identity] += 1

    results['number_of_errors_in_all_blobs'] = {i: results['number_of_individuals_badly_assigned'][i] +
                                                results['number_of_individuals_badly_interpolated'][i]
                                                for i in range(1, number_of_animals + 1)}

    return results

def get_accuracy_wrt_groundtruth_no_gaps(video, groundtruth,
                                            blobs_in_video_groundtruth,
                                            blobs_in_video = None,
                                            first_frame_first_global_fragment = None):

    check_ground_truth_consistency(blobs_in_video_groundtruth, blobs_in_video, first_frame_first_global_fragment)
    identities_dictionary_permutation = get_permutation_of_identities(first_frame_first_global_fragment,
                                                                    blobs_in_video_groundtruth, blobs_in_video)
    number_of_animals = video.number_of_animals
    if blobs_in_video is None:
        blobs_in_video = blobs_in_video_groundtruth
    results = compare_tracking_against_groundtruth_no_gaps(number_of_animals,
                                                    groundtruth, blobs_in_video_groundtruth,
                                                    blobs_in_video, identities_dictionary_permutation)

    accuracies = {}
    accuracies['individual_accuracy'] = {i : 1. - results['number_of_errors_in_all_blobs'][i] / results['number_of_blobs_per_identity'][i]
                            for i in range(1, number_of_animals + 1)}
    accuracies['accuracy'] = np.mean(list(accuracies['individual_accuracy'].values()))
    logger.info(accuracies)
    logger.info(results)
    return accuracies, results

def compare_tracking_against_groundtruth(number_of_animals, blobs_in_video_groundtruth, blobs_in_video, identities_dictionary_permutation):
    #create dictionary to store eventual corrections made by the user
    results = {}
    results['number_of_blobs_per_identity'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['sum_individual_P2'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_assigned_blobs_per_identity'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_blobs_assigned_during_accumulation_per_identity'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_blobs_after_accumulation_per_identity'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_all_blobs'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_assigned_blobs'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_blobs_assigned_during_accumulation'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_blobs_after_accumulation'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_blobs_assigned_after_accumulation'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_individual_blobs'] = 0
    results['number_of_crossing_blobs'] = 0
    results['number_of_crossings_blobs_assigned_as_individuals'] = 0
    results['frames_with_identity_errors'] = []
    results['fragment_identifiers_with_identity_errors'] = []
    results['frames_with_crossing_errors'] = []
    results['fragment_identifiers_with_crossing_errors'] = []
    results['frames_with_zeros_in_groundtruth'] = []
    results['number_of_crossing_fragments'] = 0
    results['fragments_identifiers_of_crossings'] = []

    for groundtruth_blobs_in_frame, blobs_in_frame in zip(blobs_in_video_groundtruth, blobs_in_video):

        for groundtruth_blob, blob in zip(groundtruth_blobs_in_frame,blobs_in_frame):

            if identities_dictionary_permutation is not None:
                gt_identity = identities_dictionary_permutation[groundtruth_blob.identity]
            else:
                gt_identity = groundtruth_blob.identity

            if groundtruth_blob.is_an_individual and gt_identity != -1 and not groundtruth_blob.was_a_crossing: # we are not considering crossing or failures of the model area
                results['number_of_individual_blobs'] += 1
                if gt_identity == 0:
                    results['frames_with_zeros_in_groundtruth'].append(groundtruth_blob.frame_number)
                else:
                    try:
                        if blob.assigned_identities[0] != 0 and blob.identities_corrected_closing_gaps is None: # we only consider P2 for non interpolated blobs
                            results['sum_individual_P2'][gt_identity] += blob._P2_vector[gt_identity - 1]
                    except:
                        logger.debug("P2_vector %s" %str(blob._P2_vector))
                        logger.debug("individual %s" %str(blob.is_an_individual))
                        logger.debug("fragment identifier %s" %str(blob.fragment_identifier))
                    results['number_of_blobs_per_identity'][gt_identity] += 1
                    results['number_of_assigned_blobs_per_identity'][gt_identity] += 1 if blob.assigned_identities[0] != 0 else 0
                    results['number_of_blobs_assigned_during_accumulation_per_identity'][gt_identity] += 1 if blob.used_for_training else 0
                    results['number_of_blobs_after_accumulation_per_identity'][gt_identity] += 1 if not blob.used_for_training else 0
                    if gt_identity != blob.assigned_identities[0]:
                        results['number_of_errors_in_all_blobs'][gt_identity] += 1
                        results['number_of_errors_in_blobs_after_accumulation'][gt_identity] += 1 if not blob.used_for_training else 0
                        if blob.assigned_identities[0] != 0:
                            results['number_of_errors_in_assigned_blobs'][gt_identity] += 1
                            results['number_of_errors_in_blobs_assigned_during_accumulation'][gt_identity] += 1 if blob.used_for_training else 0
                            results['number_of_errors_in_blobs_assigned_after_accumulation'][gt_identity] += 1 if not blob.used_for_training else 0
                        if blob.fragment_identifier not in results['fragment_identifiers_with_identity_errors']:
                            results['frames_with_identity_errors'].append(blob.frame_number)
                            results['fragment_identifiers_with_identity_errors'].append(blob.fragment_identifier)

            elif groundtruth_blob.is_a_crossing or gt_identity == -1:
                if blob.fragment_identifier not in results['fragments_identifiers_of_crossings']:
                    results['fragments_identifiers_of_crossings'].append(blob.fragment_identifier)
                    results['number_of_crossing_fragments'] += 1
                results['number_of_crossing_blobs'] += 1
                results['number_of_crossings_blobs_assigned_as_individuals'] += 1 if blob.is_an_individual else 0
                if blob.is_an_individual:
                    if blob.fragment_identifier not in results['fragment_identifiers_with_crossing_errors']:
                        results['frames_with_crossing_errors'].append(blob.frame_number)
                        results['fragment_identifiers_with_crossing_errors'].append(blob.fragment_identifier)

    return results

def check_ground_truth_consistency(blobs_in_video_groundtruth, blobs_in_video, first_frame_first_global_fragment):

    if first_frame_first_global_fragment is not None \
        and first_frame_first_global_fragment > len(blobs_in_video_groundtruth):
        raise ValueError('The first_frame_first_global_fragment is bigger than the length of the groundtruth video')

    if len(blobs_in_video_groundtruth) != len(blobs_in_video):
        raise ValueError('Cannot compute the accuracy from a list of blobs with different length than the blobs_in_video_groundtruth')

    for blobs_in_frame_gt, blobs_in_frame in zip(blobs_in_video_groundtruth, blobs_in_video):

        if len(blobs_in_frame) != len(blobs_in_frame_gt):
            raise ValueError('Cannot compute the accuracy form a list of blobs with different blobs per frame than the blobs_in_video_groundtruth')

def get_permutation_of_identities(first_frame_first_global_fragment, blobs_in_video_groundtruth, blobs_in_video):
    if first_frame_first_global_fragment is not None:
        groundtruth_identities_in_first_frame = [blob.identity for blob in blobs_in_video_groundtruth[first_frame_first_global_fragment]]
        identities_in_first_frame = [blob.identity for blob in blobs_in_video[first_frame_first_global_fragment]]
        logger.debug('groundtruth identities in first frame %s' %str(groundtruth_identities_in_first_frame))
        logger.debug('identities in first frame %s' %str(identities_in_first_frame))

        identities_dictionary_permutation = {groundtruth_identity: identity
                                                for identity, groundtruth_identity in zip(identities_in_first_frame, groundtruth_identities_in_first_frame)}
    else:
        identities_dictionary_permutation = None

    return identities_dictionary_permutation

def get_accuracy_wrt_groundtruth(video, blobs_in_video_groundtruth, blobs_in_video = None, first_frame_first_global_fragment = None):
    check_ground_truth_consistency(blobs_in_video_groundtruth, blobs_in_video, first_frame_first_global_fragment)
    identities_dictionary_permutation = get_permutation_of_identities(first_frame_first_global_fragment, blobs_in_video_groundtruth, blobs_in_video)

    number_of_animals = video.number_of_animals
    if blobs_in_video is None:
        blobs_in_video = blobs_in_video_groundtruth
    results = compare_tracking_against_groundtruth(number_of_animals, blobs_in_video_groundtruth, blobs_in_video, identities_dictionary_permutation)
    if len(results['frames_with_zeros_in_groundtruth']) == 0:
        accuracies = {}
        accuracies['percentage_of_unoccluded_images'] = results['number_of_individual_blobs'] / (results['number_of_individual_blobs'] + results['number_of_crossing_blobs'])
        accuracies['individual_P2_in_validated_part'] = {i : results['sum_individual_P2'][i] / results['number_of_blobs_per_identity'][i]
                                for i in range(1, number_of_animals + 1)}
        accuracies['mean_individual_P2_in_validated_part'] = np.sum(list(results['sum_individual_P2'].values())) / np.sum(list(results['number_of_blobs_per_identity'].values()))
        accuracies['individual_accuracy'] = {i : 1 - results['number_of_errors_in_all_blobs'][i] / results['number_of_blobs_per_identity'][i]
                                for i in range(1, number_of_animals + 1)}
        accuracies['accuracy'] = 1. - np.sum(list(results['number_of_errors_in_all_blobs'].values())) / np.sum(list(results['number_of_blobs_per_identity'].values()))
        accuracies['individual_accuracy_assigned'] = {i : 1 - results['number_of_errors_in_assigned_blobs'][i] / results['number_of_assigned_blobs_per_identity'][i]
                                        for i in range(1, number_of_animals + 1)}
        accuracies['accuracy_assigned'] = 1. - np.sum(list(results['number_of_errors_in_assigned_blobs'].values())) / np.sum(list(results['number_of_assigned_blobs_per_identity'].values()))
        accuracies['individual_accuracy_in_accumulation'] = {i : 1 - results['number_of_errors_in_blobs_assigned_during_accumulation'][i] / results['number_of_blobs_assigned_during_accumulation_per_identity'][i]
                                for i in range(1, number_of_animals + 1)}
        accuracies['accuracy_in_accumulation'] = 1. - np.sum(list(results['number_of_errors_in_blobs_assigned_during_accumulation'].values())) / np.sum(list(results['number_of_blobs_assigned_during_accumulation_per_identity'].values()))
        accuracies['individual_accuracy_after_accumulation'] = {}
        for i in range(1, number_of_animals + 1):
            if results['number_of_blobs_after_accumulation_per_identity'][i] != 0:
                accuracies['individual_accuracy_after_accumulation'][i] = 1 - results['number_of_errors_in_blobs_after_accumulation'][i] / results['number_of_blobs_after_accumulation_per_identity'][i]
            else:
                accuracies['individual_accuracy_after_accumulation'][i] = None
        if np.sum(list(results['number_of_blobs_after_accumulation_per_identity'].values())) != 0:
            accuracies['accuracy_after_accumulation'] = 1. - np.sum(list(results['number_of_errors_in_blobs_after_accumulation'].values())) / np.sum(list(results['number_of_blobs_after_accumulation_per_identity'].values()))
        else:
            accuracies['accuracy_after_accumulation'] = None
        if results['number_of_crossing_blobs'] != 0:
            accuracies['crossing_detector_accuracy'] = 1. - results['number_of_crossings_blobs_assigned_as_individuals'] / results['number_of_crossing_blobs']
        else:
            accuracies['crossing_detector_accuracy'] = None
        logger.info("accuracies %s" %str(accuracies))
        logger.info("number of crossing fragments in ground truth interval: %i" %results['number_of_crossing_fragments'])
        logger.info("number of crossing blobs in ground truth interval: %i" %results['number_of_crossing_blobs'])
        return accuracies, results

    else:
        logger.info("there are fish with 0 identity in frame %s" %str(results['frames_with_zeros_in_groundtruth']))
        return None, results

def compute_and_save_session_accuracy_wrt_groundtruth(video, groundtruth_type = None):
    logger.info("loading list_of_blobs")
    if groundtruth_type == 'normal':
        list_of_blobs = ListOfBlobs.load(video, video.blobs_path)
    elif groundtruth_type == 'interpolated':
        list_of_blobs = ListOfBlobs.load(video, video.blobs_path_interpolated)
    elif groundtruth_type == 'no_gaps':
        list_of_blobs = ListOfBlobs.load(video, video.blobs_no_gaps_path)
    #select ground truth file
    logger.info("loading groundtruth")
    if groundtruth_type == 'normal' or groundtruth_type == 'interpolated':
        groundtruth_path = os.path.join(video.video_folder,'_groundtruth.npy')
    elif groundtruth_type == 'no_gaps':
        groundtruth_path = os.path.join(video.video_folder,'_groundtruth_with_crossing_identified.npy')
    groundtruth = np.load(groundtruth_path, allow_pickle=True).item()
    blobs_in_video_groundtruth = groundtruth.blobs_in_video[groundtruth.start:groundtruth.end]
    blobs_in_video = list_of_blobs.blobs_in_video[groundtruth.start:groundtruth.end]
    logger.info("computing groundtruth")
    if groundtruth_type == 'normal' or groundtruth_type == 'interpolated':
        accuracies, results = get_accuracy_wrt_groundtruth(video, blobs_in_video_groundtruth, blobs_in_video)
    elif groundtruth_type == 'no_gaps':
        accuracies, results = get_accuracy_wrt_groundtruth_no_gaps(video, groundtruth, blobs_in_video_groundtruth, blobs_in_video)
    if accuracies is not None:
        logger.info("saving accuracies in video")
        video.gt_start_end = (groundtruth.start,groundtruth.end)
        if groundtruth_type == 'normal':
            video.gt_accuracy = accuracies
            video.gt_results = results
        elif groundtruth_type == 'interpolated':
            video.gt_accuracy_interpolated = accuracies
            video.gt_results_interpolated = results
        elif groundtruth_type == 'no_gaps':
            video.gt_accuracy_no_gaps = accuracies
            video.gt_results_no_gaps = results
        video.save()


if __name__ == '__main__':
    from idtrackerai.utils.GUI_utils import selectDir

    groundtruth_type = sys.argv[1]
    #select blobs_in_video list tracked to compare against ground truth
    session_path = selectDir('./') #select path to video
    video_object_path = os.path.join(session_path,'video_object.npy')
    logger.info("loading video object")
    video = np.load(video_object_path, allow_pickle=True).item(0)
    compute_and_save_session_accuracy_wrt_groundtruth(video, groundtruth_type)



import numpy as np
import time

from confapp import conf

from idtrackerai.list_of_blobs import ListOfBlobs
from idtrackerai.preprocessing.segmentation import segment
from idtrackerai.crossing_detector import detect_crossings

from idtrackerai.list_of_global_fragments import ListOfGlobalFragments, create_list_of_global_fragments
from idtrackerai.list_of_fragments        import ListOfFragments, create_list_of_fragments

if not hasattr(conf,'PYFORMS_MODE'):
    from kivy.logger import Logger as logger
else:
    import logging; logger = logging.getLogger(__name__)


class PreprocessingPreviewAPI(object):

    def __init__(self, chosen_video=None, **kwargs):

        #: Chosen_Video: ?
        self.chosen_video  = chosen_video
        #: int: Used to filter the image to find the blobs
        self.min_threshold = chosen_video.video.min_threshold if chosen_video.video is not None else 0
        #: int: Used to filter the image to find the blobs
        self.max_threshold = chosen_video.video.max_threshold if chosen_video.video is not None else 135
        #: int: Minimum area of a blob
        self.min_area      = chosen_video.video.min_area if chosen_video.video is not None else 150
        #: int: Maximum area of a blob
        self.max_area      = chosen_video.video.max_area if chosen_video.video is not None else 60000

        #: float: ?
        self.resolution_reduction = chosen_video.video.resolution_reduction if chosen_video.video is not None else 1
        #: int: Number of animals to track
        self.number_of_animals    = chosen_video.video.number_of_animals if chosen_video.video.number_of_animals is not None else 1
        #: ListOfFragments: List of fragments ( blobs paths before crossing )
        self.list_of_fragments    = None
        #: list(GlobalFragment): ?
        self.list_of_global_fragments = None
        #: ?: ?
        self.crossing_detector_trainer= None
        #: boolean: ?
        self.resegmentation_step_finished = True
        #: list(int): Indexes of the frames with more blobs than animals to track
        self.frames_with_more_blobs_than_animals = None


    def init_preview(self):

        logger.debug("init_preview")
        self.init_preproc_parameters()

        self.bkg = self.chosen_video.video.bkg
        self.ROI = self.chosen_video.video.ROI if self.chosen_video.video.ROI is not None else np.ones((self.chosen_video.video.original_height, self.chosen_video.video.original_width) ,dtype='uint8') * 255

        logger.debug("init_segment_zero")
        self.init_segment_zero()


    def init_segment_zero(self):
        self.currentSegment = 0
        self.areas_plotted  = False
        self.number_of_detected_blobs = [0]


    def init_preproc_parameters(self):

        logger.debug("init_preproc_parameters")

        if self.chosen_video.old_video is not None and self.chosen_video.old_video._has_been_preprocessed == True:

            self.max_threshold = self.chosen_video.old_video.max_threshold
            self.min_threshold = self.chosen_video.old_video.min_threshold
            self.min_area      = self.chosen_video.old_video.min_area
            self.max_area      = self.chosen_video.old_video.max_area
            self.resolution_reduction = self.chosen_video.old_video.resolution_reduction
            self.number_of_animals    = self.chosen_video.old_video.number_of_animals
            self.chosen_video.video.resolution_reduction = self.chosen_video.old_video.resolution_reduction

        else:

            self.max_threshold = conf.MAX_THRESHOLD_DEFAULT
            self.min_threshold = conf.MIN_THRESHOLD_DEFAULT
            self.min_area      = conf.MIN_AREA_DEFAULT
            self.max_area      = conf.MAX_AREA_DEFAULT
            self.resolution_reduction = self.chosen_video.video.resolution_reduction
            if self.chosen_video.video._original_ROI is None:
                self.chosen_video.video._original_ROI = np.ones( (self.chosen_video.video.original_height, self.chosen_video.video.original_width), dtype='uint8') * 255


    def compute_list_of_blobs(self, *args):
        self.blobs = segment(self.chosen_video.video)
        self.chosen_video.list_of_blobs = ListOfBlobs(blobs_in_video = self.blobs)
        self.chosen_video.video.create_preprocessing_folder()


    def segment(self, min_threshold, max_threshold, min_area, max_area):
        self.chosen_video.video._segmentation_time = time.time()
        logger.debug("segment")
        self.chosen_video.video._max_threshold = max_threshold
        self.chosen_video.video._min_threshold = min_threshold
        self.chosen_video.video._min_area = min_area
        self.chosen_video.video._max_area = max_area
        logger.debug("segment 1")
        self.chosen_video.video.resolution_reduction = self.resolution_reduction
        logger.debug("segment 2")
        self.chosen_video.video.save()


    def check_segmentation_consistency(self):
        """
        :return: True if the segmentation is consistence with the number of animals, otherwise return False
        """
        self.chosen_video.video.frames_with_more_blobs_than_animals, self.chosen_video.video._maximum_number_of_blobs = \
            self.chosen_video.list_of_blobs.check_maximal_number_of_blob(
                self.chosen_video.video.number_of_animals,
                return_maximum_number_of_blobs=True
            )
        self.frames_with_more_blobs_than_animals = self.chosen_video.video.frames_with_more_blobs_than_animals

        """
        #This call is used in the GUI to re-segment the image in the case the tracking returned more blobs than expected
        #In the case of the API this is not necessary
        if len(self.frames_with_more_blobs_than_animals) > 0 and (check_segmentation_consistency or self.chosen_video.video.number_of_animals == 1):
            self.resegmentation_step_finished = True

            if resegment: self.resegmentation()
        """
        return len(self.chosen_video.video.frames_with_more_blobs_than_animals)==0



    def save_list_of_blobs_segmented(self):
        self.chosen_video.video._has_been_segmented = True

        if len(self.chosen_video.list_of_blobs.blobs_in_video[-1]) == 0:
            self.chosen_video.list_of_blobs.blobs_in_video   = self.chosen_video.list_of_blobs.blobs_in_video[:-1]
            self.chosen_video.list_of_blobs.number_of_frames = len(self.chosen_video.list_of_blobs.blobs_in_video)
            self.chosen_video.video._number_of_frames        = self.chosen_video.list_of_blobs.number_of_frames

        self.chosen_video.video.save()
        # NOTE: The name of this functions should be changed. We do not need to save
        # the list_of_blobs after segmentation as we are not restoring the segmentation anymore
        # Even if we restore, we restore the whole preprocessing, not only the segmentation

        # self.chosen_video.list_of_blobs.save(
        #     self.chosen_video.video,
        #     self.chosen_video.video.blobs_path_segmented,
        #     number_of_chunks = self.chosen_video.video.number_of_frames
        # )

        self.chosen_video.video._segmentation_time = time.time() - self.chosen_video.video.segmentation_time


    def model_area_and_crossing_detector(self):
        self.chosen_video.video._crossing_detector_time = time.time()

        self.chosen_video.video._model_area, self.chosen_video.video._median_body_length = self.chosen_video.list_of_blobs.compute_model_area_and_body_length(
            self.chosen_video.video.number_of_animals
        )
        self.chosen_video.video.compute_identification_image_size(self.chosen_video.video.median_body_length)

        if not self.chosen_video.list_of_blobs.blobs_are_connected:
            self.chosen_video.list_of_blobs.compute_overlapping_between_subsequent_frames()


    def train_and_apply_crossing_detector(self):

        if self.chosen_video.video.number_of_animals != 1:

            self.crossing_detector_trainer = detect_crossings(
                self.chosen_video.list_of_blobs,
                self.chosen_video.video,
                self.chosen_video.video.model_area,
                use_network = True,
                return_store_objects = True,
                plot_flag = conf.PLOT_CROSSING_DETECTOR
            )
        else:

            self.chosen_video.list_of_blob = detect_crossings(
                self.chosen_video.list_of_blobs,
                self.chosen_video.video,
                self.chosen_video.video.model_area,
                use_network = False,
                return_store_objects = False,
                plot_flag = conf.PLOT_CROSSING_DETECTOR
            )
            self.chosen_video.list_of_blob.save(
                self.chosen_video.video,
                self.chosen_video.video.blobs_path_segmented,
                number_of_chunks = self.chosen_video.video.number_of_frames
            )


    def generate_list_of_fragments_and_global_fragments(self):
        self.chosen_video.video._fragmentation_time = time.time()

        if self.chosen_video.video.number_of_animals != 1:
            self.chosen_video.list_of_blobs.compute_overlapping_between_subsequent_frames()
            self.chosen_video.list_of_blobs.compute_fragment_identifier_and_blob_index(max(self.chosen_video.video.number_of_animals, self.chosen_video.video.maximum_number_of_blobs))
            self.chosen_video.list_of_blobs.compute_crossing_fragment_identifier()
            fragments = create_list_of_fragments(self.chosen_video.list_of_blobs.blobs_in_video,
                                                self.chosen_video.video.number_of_animals)
            self.list_of_fragments = ListOfFragments(fragments,
                                                     self.chosen_video.video.identification_images_file_path)
            self.chosen_video.video._fragment_identifier_to_index = self.list_of_fragments.get_fragment_identifier_to_index_list()
            global_fragments = create_list_of_global_fragments(self.chosen_video.list_of_blobs.blobs_in_video,
                                                                self.list_of_fragments.fragments,
                                                                self.chosen_video.video.number_of_animals)
            self.list_of_global_fragments = ListOfGlobalFragments(global_fragments)
            self.chosen_video.video.number_of_global_fragments = self.list_of_global_fragments.number_of_global_fragments
            self.list_of_global_fragments.filter_candidates_global_fragments_for_accumulation()
            self.chosen_video.video.number_of_global_fragments_candidates_for_accumulation = self.list_of_global_fragments.number_of_global_fragments
            #XXX I skip the fit of the gamma ...
            self.list_of_global_fragments.relink_fragments_to_global_fragments(self.list_of_fragments.fragments)
            self.list_of_global_fragments.compute_maximum_number_of_images()
            self.chosen_video.video._maximum_number_of_images_in_global_fragments = self.list_of_global_fragments.maximum_number_of_images
            self.list_of_fragments.get_accumulable_individual_fragments_identifiers(self.list_of_global_fragments)
            self.list_of_fragments.get_not_accumulable_individual_fragments_identifiers(self.list_of_global_fragments)
            self.list_of_fragments.set_fragments_as_accumulable_or_not_accumulable()
            self.chosen_video.video._number_of_unique_images_in_global_fragments = self.list_of_fragments.compute_total_number_of_images_in_global_fragments()
            self.list_of_fragments.save(self.chosen_video.video.fragments_path)
            self.chosen_video.list_of_fragments = self.list_of_fragments
        else:
            self.chosen_video.video._number_of_unique_images_in_global_fragments = None
            self.chosen_video.video._maximum_number_of_images_in_global_fragments = None
        self.chosen_video.video._has_been_preprocessed = True
        self.chosen_video.list_of_blobs.save(self.chosen_video.video, self.chosen_video.video.blobs_path, number_of_chunks = self.chosen_video.video.number_of_frames)
        if self.chosen_video.video.number_of_animals != 1:
            self.list_of_global_fragments.save(self.chosen_video.video.global_fragments_path, self.list_of_fragments.fragments)
            self.chosen_video.list_of_global_fragments = self.list_of_global_fragments
        self.chosen_video.video._fragmentation_time = time.time() - self.chosen_video.video.fragmentation_time
        self.chosen_video.video.save()



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys
import os

MIN_FLOAT = sys.float_info[3]
MAX_FLOAT = sys.float_info[0]

IDTRACKERAI_FOLDER = os.path.dirname(os.path.abspath(__file__))
TEST_VIDEO_URL='https://drive.google.com/uc?id=1uBOEMGxrOed8du7J9Rt-dlXdqOyhCpMC'

PROCESSES = ['preprocessing','protocols1_and_2', 'protocol3_pretraining',
            'protocol3_accumulation', 'residual_identification',
            'post_processing']

#######################################
##########       video      ###########
#######################################
'''***AVAILABLE_VIDEO_EXTENSION***
Tested and available video extensions
'''
AVAILABLE_VIDEO_EXTENSION = ['.avi', '.AVI', '.mp4', '.MP4', '.mpg', '.MPG', '.mov', '.MOV']
'''***FRAMES_PER_EPISODE***
Number of frames per video chunk. Used to parallelise processes
'''
FRAMES_PER_EPISODE = int(os.environ.get('FRAMES_PER_EPISODE', 500))
#######################################
##########   preprocessing  ###########
#######################################
"""***NUMBER_OF_ANIMALS_DEFAULT
Number of animals to be tracked
"""
NUMBER_OF_ANIMALS_DEFAULT = int(os.environ.get('FRAMES_PER_EPISODE', 8))
"""***MIN_AREA_LOWER, MIN_AREA_UPPER***
Lower and upper bounds for the minimum area slider
"""
MIN_AREA_LOWER, MIN_AREA_UPPER = 0, 10000
"""***DEFAULT_RESOLUTION_REDUCTION***
Ratio to which the width and height are rescaled.
"""
RES_REDUCTION_DEFAULT = float(os.environ.get('RES_REDUCTION_DEFAULT', 1.0))
"""***MIN_AREA_DEFAULT, MAX_AREA_DEFAULT***
Default value for min area in preprocessing
"""
MIN_AREA_DEFAULT, MAX_AREA_DEFAULT = 150, 60000
"""***AREA_LOWER, AREA_UPPER***
Lower and upper bounds for the maximum area slider
"""
AREA_LOWER, AREA_UPPER = 0, 60000
"""***MIN_THRESHOLD, AREA_UPPER***
Lower and upper bounds for the maximum area slider
"""
MIN_THRESHOLD, MAX_THRESHOLD = 0, 255
"""***MIN_THRESHOLD_DEFAULT, MAX_THRESHOLD_DEFAULT***
Default value for min area in preprocessing
"""
MIN_THRESHOLD_DEFAULT, MAX_THRESHOLD_DEFAULT = 0, 155
"""***VEL_PERCENTILE***
percentile on the average speed of the individuals used to compute the maximal
velocity threshold
"""
VEL_PERCENTILE = int(os.environ.get('VEL_PERCENTILE', 99))
'''***STD_TOLERANCE***
Tolerance coefficient in the computation of the individual model area.
'''
STD_TOLERANCE = int(os.environ.get('STD_TOLERANCE', 4))
"""***BACKGROUND_SUBTRACTION_PERIOD***
Period used to sample the video to compute the background model
"""
BACKGROUND_SUBTRACTION_PERIOD = int(os.environ.get('BACKGROUND_SUBTRACTION_PERIOD', 500))
"""***NUMBER_OF_JOBS_FOR_BACKGROUND_SUBTRACTION***
Number of jobs used to compute the background model
"""
NUMBER_OF_JOBS_FOR_BACKGROUND_SUBTRACTION = -2 # Set None to use the default mode of the system. (see segmentation_utils.py module for details)
"""***NUMBER_OF_JOBS_FOR_SEGMENTATION***
Number of jobs used to perform the segmentation
"""
NUMBER_OF_JOBS_FOR_SEGMENTATION = -2 # Set None to use the default mode of the system. (see segmentation.py module for details)
"""***SIGMA_GAUSSIAN_BLURRING***
sigma for gaussian blurring tests
"""
SIGMA_GAUSSIAN_BLURRING = None
"""***SAVE_PIXELS***
Where to store the pixels list of every blob
DISK: pixels are stored in hdf5 files in the disk
RAM: pixels are stored in the blob object in the ram
NOT: pixels are not stored and are computed from the contour everytime
"""
SAVE_PIXELS = 'DISK' # 'RAM' or 'NONE'
"""***SAVE_SEGMENTATION_IMAGE***
Where to store the segmentation image of every blob
DISK: image is stored in hdf5 files in the disk
RAM: image is stored in the blob object in the ram
NOT: image is not stored and are computed from the contour everytime
"""
SAVE_SEGMENTATION_IMAGE = 'DISK' # 'RAM' or 'NONE'
"""***PLOT_CROSSING_DETECTOR***
"""
PLOT_CROSSING_DETECTOR=False
#######################################
#########  global fragments  ##########
#######################################
'''***MINIMUM_NUMBER_OF_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION***
Minimum number of frame to allow an individual fragment to be part of a
global one
'''
MINIMUM_NUMBER_OF_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION = int(os.environ.get('MINIMUM_NUMBER_OF_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION', 3))
#######################################
##########        CNN       ###########
#######################################
'''***PLOT_ACCUMULATION_STEPS***
'''
PLOT_ACCUMULATION_STEPS = False
'''***CNN_MODEL***
Architecture for the IDCNN network. By default the model 0 is used, which correspond
to the model of the paper. Check the different models in the module cnn_architectures.py
in the folder network and how they are map to a single value code in the module
id_CNN.py in the network/identification_model folder.
'''
CNN_MODEL = 0
'''***KNOWLEDGE_TRANSFER_FOLDER_IDCNN***
Folder for a accumulation folder with a model from another video. Note that if
the IDENTITY_TRANSFER flag is True, then the IDENTIFICATION_IMAGE_SIZE will be
taken from the knowledge_transfer_info_dict.
'''
KNOWLEDGE_TRANSFER_FOLDER_IDCNN = None
'''***IDENTITY_TRANSFER***
Bloonean
'''
IDENTITY_TRANSFER = False
"""***IDENTIFICATION_IMAGE_SIZE***
size of the identification images. Used for idmatcher.ai
"""
IDENTIFICATION_IMAGE_SIZE = None #It should be a tuple of len 3 (width, height, chanels), e.g. (46, 46, 1)
'''***USE_ADAM_OPTIMISER***
By default the identification neural network is trained with SGD. If this parameter
is set to TRUE, the network will be trained with the ADAM optimizer.
'''
USE_ADAM_OPTIMISER = False
'''***LAYERS_TO_OPTIMISE_ACCUMULATION
List of layers to be optimized. None indicates that all layers are optimizsed.
To check the name of the different layers see the module cnn_architectures.py
in the folder network.
'''
LAYERS_TO_OPTIMISE_ACCUMULATION = None # ['fully-connected1','fully_connected_pre_softmax']
LAYERS_TO_OPTIMISE_PRETRAINING = None
'''***LEARNING_RATE_IDCNN_ACCUMULATION***
'''
LEARNING_RATE_IDCNN_ACCUMULATION = float(os.environ.get('LEARNING_RATE_IDCNN_PRETRAINING', 0.005))
'''***LEARNING_RATE_IDCNN_PRETRAINING***
'''
LEARNING_RATE_IDCNN_PRETRAINING = float(os.environ.get('LEARNING_RATE_IDCNN_PRETRAINING', 0.01))
'''***LEARNING_RATE_DCD***
'''
LEARNING_RATE_DCD = float(os.environ.get('LEARNING_RATE_DCD', 0.001))
'''***KEEP_PROB_IDCNN_ACCUMULATION***
'''
KEEP_PROB_IDCNN_ACCUMULATION = float(os.environ.get('KEEP_PROB_IDCNN_ACCUMULATION', 1.))
'''***KEEP_PROB_IDCNN_PRETRAINING***
'''
KEEP_PROB_IDCNN_PRETRAINING = float(os.environ.get('KEEP_PROB_IDCNN_PRETRAINING', 1.))
'''***KEEP_PROB_DCD***
'''
KEEP_PROB_DCD = float(os.environ.get('KEEP_PROB_DCD', 1.))
'''***VALIDATION_PROPORTION***
Protortion of images used for validation in the IDCNN model
'''
VALIDATION_PROPORTION = float(os.environ.get('VALIDATION_PROPORTION', .1))
'''***BATCH_SIZE_DCD, BATCH_SIZE_IDCNN***
size of the batches used to train the DCD and idCNN, respectively.
'''
BATCH_SIZE_IDCNN = int(os.environ.get('BATCH_SIZE_IDCNN', 50))
BATCH_SIZE_DCD = int(os.environ.get('BATCH_SIZE_DCD', 50))
'''***BATCH_SIZE_PREDICTIONS***
size of the batches used to get the output from the DCD and idCNN, respectively.
Remark: This is done to prevent out-of-memory error in the GPU
'''
BATCH_SIZE_PREDICTIONS_DCD = int(os.environ.get('BATCH_SIZE_PREDICTIONS_DCD', 100))
BATCH_SIZE_PREDICTIONS_IDCNN = int(os.environ.get('BATCH_SIZE_PREDICTIONS_IDCNN', 500))
'''***LEARNING_PERCENTAGE_DIFFERENCE_1 ***
Overfitting threshold during training
'''
LEARNING_PERCENTAGE_DIFFERENCE_1_DCD = float(os.environ.get('LEARNING_PERCENTAGE_DIFFERENCE_1_DCD', .005))
LEARNING_PERCENTAGE_DIFFERENCE_1_IDCNN = float(os.environ.get('LEARNING_PERCENTAGE_DIFFERENCE_1_IDCNN', .005))
'''***LEARNING_PERCENTAGE_DIFFERENCE_2
Loss plateau threshold during training
***'''
LEARNING_PERCENTAGE_DIFFERENCE_2_DCD = float(os.environ.get('LEARNING_PERCENTAGE_DIFFERENCE_2_DCD', .005))
LEARNING_PERCENTAGE_DIFFERENCE_2_IDCNN = float(os.environ.get('LEARNING_PERCENTAGE_DIFFERENCE_2_IDCNN', .005))
'''***OVERFITTING_COUNTER_THRESHOLD ***
Number of consecutive overfitting epochs in order to stop the training
'''
OVERFITTING_COUNTER_THRESHOLD_DCD = int(os.environ.get('OVERFITTING_COUNTER_THRESHOLD_DCD', 5))
OVERFITTING_COUNTER_THRESHOLD_IDCNN = int(os.environ.get('OVERFITTING_COUNTER_THRESHOLD_IDCNN', 5))
OVERFITTING_COUNTER_THRESHOLD_IDCNN_FIRST_ACCUM = int(os.environ.get('OVERFITTING_COUNTER_THRESHOLD_IDCNN_FIRST_ACCUM', 10))
'''***MAXIMUM_NUMBER_OF_EPOCHS ***
Maximum number of epochs before forcing the training to stop
'''
MAXIMUM_NUMBER_OF_EPOCHS_DCD = int(os.environ.get('MAXIMUM_NUMBER_OF_EPOCHS_DCD', 100))
MAXIMUM_NUMBER_OF_EPOCHS_IDCNN = int(os.environ.get('MAXIMUM_NUMBER_OF_EPOCHS_IDCNN', 10000))
'''***KEEP_PROB***
Default dropout in fully-connected layers if the CNN models (can be changed
when instantiating the parameter class to be init the CNN)
'''
KEEP_PROB = float(os.environ.get('KEEP_PROB', 1.0))
'''***MINIMUM_NUMBER_OF_CROSSINGS_TO_TRAIN_CROSSING_DETECTOR***
Minimum number of crossings required to train the crossing detector, otherwise
only the model area is used to distinguish crossings from individuals
'''
MINIMUM_NUMBER_OF_CROSSINGS_TO_TRAIN_CROSSING_DETECTOR = int(os.environ.get('MINIMUM_NUMBER_OF_CROSSINGS_TO_TRAIN_CROSSING_DETECTOR', 10))
#######################################
# Deep fingerprint protocols cascade  #
#######################################
'''***THRESHOLD_EARLY_STOP_ACCUMULATION***
If the total of accumulated images + images to be accumulated is above this
threshold, we stop the accumulation (holds for each deep fingerprint protocol)
'''
THRESHOLD_EARLY_STOP_ACCUMULATION = float(os.environ.get('THRESHOLD_EARLY_STOP_ACCUMULATION', .9995))
""" ***THRESHOLD_ACCEPTABLE_ACCUMULATION***
During the deep fingerprint protocol cascade we evaluate the information
contained in global fragments and, if necessary (1), we accumulate references.
An accumulation if considered acceptable if at the end of the process the
ratio (number of accumulated image) / (number of image potentially accumulable)
if equal or bigger than THRESHOLD_ACCEPTABLE_ACCUMULATION.
----
(1) THRESHOLD_EARLY_STOP_ACCUMULATION is not reached
"""
THRESHOLD_ACCEPTABLE_ACCUMULATION = float(os.environ.get('THRESHOLD_ACCEPTABLE_ACCUMULATION', .9))
"""***MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS
***"""
MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS = int(os.environ.get('MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS', 3))
'''***MAXIMAL_IMAGES_PER_ANIMAL***
Maximal number of images per indiviudal to be included in the training dataset
of the idCNN
'''
MAXIMAL_IMAGES_PER_ANIMAL = int(os.environ.get('MAXIMAL_IMAGES_PER_ANIMAL', 3000))
'''***RATIO_OLD***
Percentage of the MAXIMAL_IMAGES_PER_ANIMAL to be subsampled from the images
accumulated in the previous steps
'''
RATIO_OLD = float(os.environ.get('RATIO_OLD', 0.6))
'''***RATIO_NEW***
Percentage of the MAXIMAL_IMAGES_PER_ANIMAL to be subsampled from the images
accumulated in the current step
'''
RATIO_NEW = float(os.environ.get('RATIO_NEW', 0.4))
'''***CERTAINTY_THRESHOLD***
Minimal certainty of the assignment of an individual fragment to be considered
acceptable
'''
CERTAINTY_THRESHOLD = float(os.environ.get('CERTAINTY_THRESHOLD', .1))
'''***MAX_RATIO_OF_PRETRAINED_IMAGES***
Maximum ratio of accumulable images to stop protocol 3 (pretraining)
'''
MAX_RATIO_OF_PRETRAINED_IMAGES = float(os.environ.get('MAX_RATIO_OF_PRETRAINED_IMAGES', .95))
'''***MINIMUM_RATIO_OF_IMAGES_ACCUMULATED_GLOBALLY_TO_START_PARTIAL_ACCUMULATION***
Minimal ratio of accumulated images over accumulable images to allow the
partial accumulation strategy to start
'''
MINIMUM_RATIO_OF_IMAGES_ACCUMULATED_GLOBALLY_TO_START_PARTIAL_ACCUMULATION = float(os.environ.get('MINIMUM_RATIO_OF_IMAGES_ACCUMULATED_GLOBALLY_TO_START_PARTIAL_ACCUMULATION', .5))
#######################################
########   post-processing   ##########
#######################################
'''***FIXED_IDENTITY_THRESHOLD***
If the certainty of the assignment of an individual fragment is above this
threshold we consider the identification certain. Thus, it won't be modified
either during the final identification or post-processing
'''
FIXED_IDENTITY_THRESHOLD = float(os.environ.get('FIXED_IDENTITY_THRESHOLD', .9))
#######################################
##########   fish-specific  ###########
#######################################
'''***SMOOTH_SIGMA***
Parameter giving the standard deviation of the gaussian filtering in the
contour to calculate curvature
'''
SMOOTH_SIGMA = int(os.environ.get('SMOOTH_SIGMA', 10))
'''***HEAD_DIAMETER***
Distance between nose and base of the head
'''
HEAD_DIAMETER = int(os.environ.get('HEAD_DIAMETER', 20))
##################################################
##########   data management policies  ###########
##################################################
""" Data policy to be applied at the end of the tracking
'all': saves all the data as it is generated from tracking
'trajectories': saves only the trajectories
'validation': saves the information needed to validate the video
'knowledge_transfer': saves the information needed to perfom identity_transfer to another video
'idmatcher.ai': saves the information needed to perform identity_matching between videos
"""
DATA_POLICY = os.environ.get('DATA_POLICY', 'all')



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., De Polavieja, G.G.,
# (2018). idtracker.ai: Tracking all individuals in large collectives of unmarked animals (F.R.-F. and M.G.B. contributed equally to this work. Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)


from __future__ import absolute_import, division, print_function
import os
import sys
import cv2
import h5py
import numpy as np
import copy
from sklearn.decomposition import PCA
if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.blob")


class Blob(object):
    """ Object representing a blob (collection of pixels) segmented from a frame

    Attributes
    ----------

    frame_number : int
        Object containing all the parameters of the video.
    in_frame_index : int
        Hierarchy of the blob in the segmentation of the frame
    number_of_animals : int
        Number of animals to be tracked
    centroid : tuple
        Centroid as (x,y) in frame coordinate
    contour: list
        List of tuples (x,y) in frame coordinate of the pixels on the boundary of the blob
    area : int
        Number of pixels composing the blob
    bounding_box_in_frame_coordinates: list
        List of tuples [(x, y), (x + width, y + height)] of the bounding box rectangle enclosing the blob's pixels
    bounding_box_image: ndarray
        Image obtained by clipping the frame to the bounding box
    estimated_body_length: float
        Body length estimated from the bounding box
    image_for_identification_path: string
        Path to where the image for identification is saved
    pixels : list
        List of ravelled pixels belonging to the blob (wrt the full frame)
    _is_an_individual : bool
        If True the blob is associated to a single animal
    _is_a_crossing : bool
        If True the blob is associated to two or more animals touching
    _was_a_crossing : bool
        If True the blob has been generated by splitting a crossing in postprocessing
    _is_a_misclassified_individual : bool
        This property can be modified only by the user during validation. It identifies a blob that was by mistaken associated to a crossing by the DeepCrossingDetector
    next : list
        List of blob objects segmented in self.frame_number + 1 whose list of pixels intersect self.pixels
    previous : list
        List of blob objects segmented in self.frame_number - 1 whose list of pixels intersect self.pixels
    _fragment_identifier : int
        Unique integer identifying the fragment (built by blob overlapping) to which self belongs to
    _blob_index : int
        Hierarchy of the blob at the beggining of the core of the global fragment. Only used to plot accumulation steps
    _used_for_training : bool
        If True the image obtained from the blob has been used to train the idCNN
    _accumulation_step : int
        Accumulation step in which the image associated to the blob has been accumulated
    _generated_while_closing_the_gap : bool
        If True the blob has been generated while solving the crossings
    _user_generated_identity : int
        The identity corrected manually by the user during validation
    _identities_corrected_closing_gaps : list
        The identity given to the blob during in postprocessing
    _identity_corrected_solving_jumps : int
        The identity given to the blob while solving duplications
    _identity : int
        Identity associated to the blob
    is_identified : bool
        True if self.identity is not None
    final_identities : list
        Identities assigned to self after validation
    assigned_identities : list
        Identity assigned to self by the algorithm (ignoring eventual correction made by the user during validation)
    has_ambiguous_identity: bool
        True if during either accumulation of residual identification the blob has
        been associated with equal probability to two (or more) distinct identities
    nose_coordinates : tuple
        Coordinate of the nose of the blob (only for zebrafish)
    head_coordinates : tuple
        Coordinate of the centroid of the head of the blob (only for zebrafish)
    extreme1_coordinate : tuple

    extreme2_coordinates : tuple

    _resolution_reduction: float


    """
    def __init__(self, centroid, contour, area,
                 bounding_box_in_frame_coordinates, bounding_box_image=None,
                 bounding_box_images_path=None,
                 estimated_body_length=None, pixels=None,
                 number_of_animals=None, frame_number=None,
                 frame_number_in_video_path=None,
                 in_frame_index=None, pixels_path=None,
                 video_height=None, video_width=None,
                 video_path=None, pixels_are_from_eroded_blob=False,
                 has_eroded_pixels=False,
                 resolution_reduction=1.):
        self.frame_number = frame_number
        self.in_frame_index = in_frame_index
        self.number_of_animals = number_of_animals
        self.centroid = centroid
        self.contour = contour # openCV contour [[[x1,y1]],[[x2,y2]],...,[[xn,yn]]]
        self.area = area # int: number of pixels in the blob
        self.bounding_box_in_frame_coordinates = bounding_box_in_frame_coordinates #tuple of tuples: ((x1,y1),(x2,y2)) (top-left corner, bottom-right corner) in pixels
        self._bounding_box_image = bounding_box_image
        self.bounding_box_images_path = bounding_box_images_path # path to where the bounding_box_image is saved
        self.estimated_body_length = estimated_body_length
        self.image_for_identification_path = None # path where the image for identification is stored
        self.identification_image_index = None
        ##
        self._pixels_path = pixels_path
        self._pixels = pixels
        self._eroded_pixels = None
        self.video_height = video_height
        self.video_width = video_width
        ##
        self._is_an_individual = False
        self._is_a_crossing = False
        self._was_a_crossing = False
        self._is_a_misclassified_individual = False
        self.next = [] # next blob object overlapping in pixels with current blob object
        self.previous = [] # previous blob object overlapping in pixels with the current blob object
        self._fragment_identifier = None # identity in individual fragment after fragmentation
        self._blob_index = None # index of the blob to plot the individual fragments
        self._used_for_training = None
        self._accumulation_step = None
        self._generated_while_closing_the_gap = False
        self._user_generated_identities = None
        self._user_generated_centroids = None
        self.interpolated_centroids = None
        self._identities_corrected_closing_gaps = None
        self._identity_corrected_solving_jumps = None
        self._identity = None
        self.video_path = video_path
        self.frame_number_in_video_path = frame_number_in_video_path
        self.pixels_are_from_eroded_blob = pixels_are_from_eroded_blob
        self.has_eroded_pixels = False
        self._resolution_reduction = resolution_reduction

    @property
    def bounding_box_image(self):
        if self._bounding_box_image is not None:
            return self._bounding_box_image
        elif self.bounding_box_images_path is not None and os.path.isfile(self.bounding_box_images_path):
            with h5py.File(self.bounding_box_images_path, 'r') as f:
                return f[str(self.frame_number) + '-' + str(self.in_frame_index)][:]
        else:
            cap = cv2.VideoCapture(self.video_path)
            cap.set(1, self.frame_number_in_video_path)
            ret, frame = cap.read()
            bb = self.bounding_box_in_frame_coordinates
            return frame[bb[0][1]:bb[1][1], bb[0][0]:bb[1][0], 0]


    @property
    def pixels(self):
        if self._pixels is not None:
            return self._pixels
        elif self._pixels_path is not None and os.path.isfile(self._pixels_path):
            with h5py.File(self._pixels_path, 'r') as f:
                if not self.pixels_are_from_eroded_blob:
                    dataset_name = str(self.frame_number) + '-' + str(self.in_frame_index)
                else:
                    dataset_name = str(self.frame_number) + '-' + str(self.in_frame_index) + '-eroded'
                return f[dataset_name][:]
        else:
            cimg = np.zeros((self.video_height, self.video_width))
            cv2.drawContours(cimg, [self.contour], -1, color=255, thickness=-1)
            pts = np.where(cimg == 255)
            pixels = np.asarray(list(zip(pts[0], pts[1])))
            pixels = np.ravel_multi_index([pixels[:, 0], pixels[:, 1]],
                                          (self.video_height, self.video_width))
            return pixels

    @property
    def eroded_pixels(self):
        if self._eroded_pixels is not None:
            return self._pixels
        elif self._pixels_path is not None and os.path.isfile(self._pixels_path):
            with h5py.File(self._pixels_path, 'r') as f:
                return f[str(self.frame_number) + '-' + str(self.in_frame_index) + '-eroded'][:]
        else:
            cimg = np.zeros((self.video_height, self.video_width))
            cv2.drawContours(cimg, [self.contour], -1, color=255, thickness=-1)
            pts = np.where(cimg == 255)
            pixels = np.asarray(list(zip(pts[0], pts[1])))
            pixels = np.ravel_multi_index([pixels[:, 0], pixels[:, 1]],
                                          (self.video_height, self.video_width))
            return pixels

    @eroded_pixels.setter
    def eroded_pixels(self, eroded_pixels):
        if self._pixels_path is not None:
            with h5py.File(self._pixels_path, 'a') as f:
                dataset_name = str(self.frame_number) + '-' + str(self.in_frame_index) + '-eroded'
                if dataset_name in f:
                    del f[dataset_name]
                f.create_dataset(dataset_name, data=eroded_pixels)
                self.has_eroded_pixels = True
        else:
            self._eroded_pixels = eroded_pixels

    @property
    def fragment_identifier(self):
        return self._fragment_identifier

    @property
    def is_an_individual(self):
        return self._is_an_individual

    @property
    def is_a_crossing(self):
        return self._is_a_crossing

    @property
    def was_a_crossing(self):
        return self._was_a_crossing

    def check_for_multiple_next_or_previous(self, direction = None):
        """Return True if self has multiple blobs in its past or future overlapping
        history of the blob

        Parameters
        ----------
        direction : str
            "previous" or "next". If "previous" the past overlapping history will
            be checked in order to find out if the blob will split in the past.
            Symmetrically, if "next" the future overlapping history of the blob
            will be checked

        Returns
        -------
        Bool
            If True the blob splits into two or multiple overlapping blobs in its
            "past" or "future" history, depending on the parameter "direction"

        """
        opposite_direction = 'next' if direction == 'previous' else 'previous'
        current = getattr(self, direction)[0]

        while len(getattr(current, direction)) == 1 :

            current = getattr(current, direction)[0]
            if len(getattr(current, direction)) > 1:# or\
                return True
                break

        return False

    def is_a_sure_individual(self):
        """A blob marked as a sure individual will be used to train the Deep
        Crossing Detector (the artificial neural network used to disriminate
        images associated with individual from ones associated to crossings).

        Returns
        -------
        Bool
            Blob is a sure individual if:

            * it overlaps with one and only one blob in both the immediate past and future frames;
            * it never splits in both its past and future overlapping history


        """
        if self.is_an_individual and len(self.previous) == 1 \
            and len(self.next) == 1 and len(self.next[0].previous) == 1 and\
            len(self.previous[0].next) == 1:
            has_multiple_previous = self.check_for_multiple_next_or_previous('previous')
            has_multiple_next = self.check_for_multiple_next_or_previous('next')
            if not has_multiple_previous and not has_multiple_next:
                return True
        else:
            return False

    def is_a_sure_crossing(self):
        """A blob marked as a sure crossing will be used to train the Deep
        Crossing Detector (the artificial neural network used to disriminate
        images associated with individual from ones associated to crossings).

        Returns
        -------
        Bool
            Blob is a sure crossing if:

            * it overlaps with one and only one blob in both the immediate past and future frames;
            * it splits in both its past and future overlapping history

        """
        if self.is_a_crossing and (len(self.previous) > 1 or len(self.next) > 1):
            return True
        elif self.is_a_crossing and len(self.previous) == 1 and len(self.next) == 1:
            has_multiple_previous = self.check_for_multiple_next_or_previous('previous')
            has_multiple_next = self.check_for_multiple_next_or_previous('next')
            if has_multiple_previous and has_multiple_next:
                return True
        else:
            return False

    @property
    def has_ambiguous_identity(self):
        return self.is_an_individual_in_a_fragment and self.identity is list

    def overlaps_with(self, other):
        """Given a second blob object, checks if the lists of pixels of the two
        blobs intersect

        Parameters
        ----------
        other : <Blob object>
            An instance of the class Blob

        Returns
        -------
        Bool
            True if the lists of pixels have non-empty intersection
        """
        overlaps = False
        intersection = np.intersect1d(self.pixels, other.pixels,
                                      assume_unique=True)
        if len(intersection) > 0:
            overlaps = True
        return overlaps

    def now_points_to(self, other):
        """Given two consecutive blob objects updates their respective overlapping
        histories

        Parameters
        ----------
        other : <Blob object>
            An instance of the class Blob
        """
        self.next.append(other)
        other.previous.append(self)

    def squared_distance_to(self, other):
        """Returns the squared distance from the centroid of self to the centroid
        of other

        Parameters
        ----------
        other : <Blob object> or tuple
            An instance of the class Blob or a tuple (x,y)

        Returns
        -------
        float
            Squared distance between centroids

        """
        if isinstance(other, Blob):
            return np.sum((np.asarray(self.centroid) - np.asarray(other.centroid))**2)
        elif isinstance(other, (tuple, list, np.ndarray)):
            return np.sum((np.asarray(self.centroid) - np.asarray(other))**2)

    def distance_from_countour_to(self, point):
        """Returns the distance between the point passed as input and the closes
        point belonging to the contour of the blob.

        Parameters
        ----------
        point : tuple
            (x,y)

        Returns
        -------
        float
            :math:`\min_{c\in \mbox{ blob.contour}}(d(c, point))`, where :math:`d`
            is the Euclidean distance.

        """
        return np.abs(cv2.pointPolygonTest(self.contour, point, True))

    @property
    def resolution_reduction(self):
        if hasattr(self, '_resolution_reduction'):
            return self._resolution_reduction
        else:
            return 1.0

    @property
    def used_for_training(self):
        return self._used_for_training

    @property
    def accumulation_step(self):
        return self._accumulation_step

    @property
    def is_in_a_fragment(self):
        return len(self.previous) == len(self.next) == 1

    @property
    def is_an_individual_in_a_fragment(self):
        return self.is_an_individual and self.is_in_a_fragment

    @property
    def is_a_misclassified_individual(self):
        return self._is_a_misclassified_individual

    @property
    def blob_index(self):
        return self._blob_index

    @blob_index.setter
    def blob_index(self, new_blob_index):
        if self.is_an_individual_in_a_fragment:
            self._blob_index = new_blob_index

    # @property
    # def image_for_identification(self):
    #     if self.is_an_individual:
    #         with h5py.File(self.image_for_identification_path, 'r') as f:
    #             return f['identification_images'][self.identification_image_index]
    #     else:
    #         return None

    @property
    def nose_coordinates(self):
        return self._nose_coordinates

    @property
    def head_coordinates(self):
        return self._head_coordinates

    @property
    def extreme1_coordinate(self):
        return self._extreme1_coordinates

    @property
    def extreme2_coordinates(self):
        return self._extreme2_coordinates

    def in_a_global_fragment_core(self, blobs_in_frame):
        """A blob in a frame is in the core of a global fragment if in
        that frame there are as many blobs as number of animals to track

        Parameters
        ----------
        blobs_in_frame : list
            List of Blob objects representing the animals segmented in the frame
            self.frame_number

        Returns
        -------
        Bool
            True if the blob is in the core of a global fragment

        """
        return len(blobs_in_frame) == self.number_of_animals

    @property
    def identity(self):
        return self._identity

    @property
    def user_generated_identities(self):
        return self._user_generated_identities

    @property
    def identity_corrected_solving_jumps(self):
        return self._identity_corrected_solving_jumps

    @property
    def identities_corrected_closing_gaps(self):
        return self._identities_corrected_closing_gaps

    @property
    def assigned_identities(self):
        if self.identities_corrected_closing_gaps is not None:
            return self.identities_corrected_closing_gaps
        elif self.identity_corrected_solving_jumps is not None:
            return [self.identity_corrected_solving_jumps]
        return [self.identity]

    @property
    def final_identities(self):
        """
        :return: Return a list of the final centroids.
        """
        if self.user_generated_identities is not None:
            # Note that sometimes len(user_generated_identities) > len(assigned_identities)
            final_identities = []
            for i, user_generated_identity in enumerate(self.user_generated_identities):
                if user_generated_identity is None:
                    final_identities.append(self.assigned_identities[i])
                elif user_generated_identity > 0:
                    final_identities.append(user_generated_identity)

            return final_identities
        return self.assigned_identities

    @property
    def is_identified(self):
        return self._identity is not None

    @property
    def generated_while_closing_the_gap(self):
        return self._generated_while_closing_the_gap

    def compute_overlapping_with_previous_blob(self):
        number_of_previous_blobs = len(self.previous)
        if number_of_previous_blobs == 1:
            self.non_shared_information_with_previous = 1. - len(np.intersect1d(self.pixels, self.previous[0].pixels)) / np.mean([len(self.pixels), len(self.previous[0].pixels)])
            if self.non_shared_information_with_previous is np.nan:
                logger.debug("intersection both blobs %s" %str(len(np.intersect1d(self.pixels, self.previous[0].pixels))))
                logger.debug("mean pixels both blobs %s" %str(np.mean([len(self.pixels), len(self.previous[0].pixels)])))
                raise ValueError("non_shared_information_with_previous is nan")

    def apply_model_area(self, video, number_of_animals, model_area,
                         identification_image_size, number_of_blobs):
        """Classify self as a crossing or individual blob according to its area

        Parameters
        ----------
        video : <Video object>
            Object containing all the parameters of the video.
        number_of_animals : int
            Number of animals to be tracked
        model_area : function
            Model of the area blobs representing individual animals
        identification_image_size : tuple
            Shape of the images used for the identification
        number_of_blobs : int
            Number of blobs segmented in the frame self.frame_number

        """
        if model_area(self.area) or number_of_blobs == number_of_animals or video.number_of_animals == 1: #Checks if area is compatible with the model area we built
            self.set_image_for_identification(video)
            self._is_an_individual = True
        else:
            self._is_a_crossing = True

    def set_image_for_identification(self, video):
        """Set the image that will be used to identitfy the animal with the idCNN

        Parameters
        ----------
        video : <Video object>
            Object containing all the parameters of the video.

        """
        image_for_identification, \
            self._extreme1_coordinates, \
            self._extreme2_coordinates, _ = self.get_image_for_identification(video)

        # For RAM optimization
        with h5py.File(video.identification_images_file_path, 'a') as f:
            dset = f['identification_images']
            i = dset.shape[0]
            dset.resize((i + 1,
                         image_for_identification.shape[1],
                         image_for_identification.shape[1]))
            dset[i, ...] = image_for_identification
        self.identification_image_index = i

    def get_image_for_identification(self, video, folder_to_save_for_paper_figure = '', image_size = None):
        """Compute the image that will be used to identify the animal with the idCNN

        Parameters
        ----------
        video : <Video object>
            Object containing all the parameters of the video.
        folder_to_save_for_paper_figure : str
            Path to the folder in which the image will be saved. If '' the image
            will not be saved.
        image_size : tuple
            Shape of the image

        Returns
        -------
        ndarray
            Stadardised image

        """
        if image_size is None:
            image_size = video.identification_image_size[0]

        return self._get_image_for_identification(video.height, video.width,
                                                self.bounding_box_image, self.pixels,
                                                self.bounding_box_in_frame_coordinates,
                                                image_size,
                                                folder_to_save_for_paper_figure=folder_to_save_for_paper_figure)


    @staticmethod
    def _get_image_for_identification(height, width, bounding_box_image, pixels, bounding_box_in_frame_coordinates, identification_image_size, folder_to_save_for_paper_figure = ''):
        """Short summary.

        Parameters
        ----------
        height : int
            Frame height
        width : int
            Frame width
        bounding_box_image : ndarray
            Images cropped from the frame by considering the bounding box associated to a blob
        pixels : list
            List of pixels associated to a blob
        bounding_box_in_frame_coordinates : list
            [(x, y), (x + bounding_box_width, y + bounding_box_height)]
        identification_image_size : tuple
            shape of the identification image
        folder_to_save_for_paper_figure : str
            folder to save the images for identification

        Returns
        -------
        ndarray
            Standardised image

        """
        if folder_to_save_for_paper_figure:
            save_preprocessing_step_image(bounding_box_image/255, folder_to_save_for_paper_figure, name = '0_bounding_box_image', min_max = [0, 1])
        bounding_box_image = remove_background_pixels(height, width, bounding_box_image, pixels, bounding_box_in_frame_coordinates, folder_to_save_for_paper_figure)
        pca = PCA()
        pxs = np.unravel_index(pixels,(height,width))
        pxs1 = np.asarray(list(zip(pxs[1],pxs[0])))
        pca.fit(pxs1)
        rot_ang = np.arctan(pca.components_[0][1]/pca.components_[0][0])*180/np.pi + 45 # we substract 45 so that the fish is aligned in the diagonal. This way we have smaller frames
        center = (pca.mean_[0], pca.mean_[1])
        center = full2miniframe(center, bounding_box_in_frame_coordinates)
        center = np.array([int(center[0]), int(center[1])])

        if folder_to_save_for_paper_figure:
            pxs_for_plot = np.array(pxs).T
            pxs_for_plot = np.array([pxs_for_plot[:, 0] - bounding_box_in_frame_coordinates[0][1], pxs_for_plot[:, 1] - bounding_box_in_frame_coordinates[0][0]])
            temp_image = np.zeros_like(bounding_box_image).astype('uint8')
            temp_image[pxs_for_plot[0,:], pxs_for_plot[1,:]] = 255
            slope = np.tan((rot_ang - 45) * np.pi / 180)
            X = [0, temp_image.shape[1]]
            Y = [-slope * center[0] + center[1], slope * (X[1] - center[0]) + center[1]]
            save_preprocessing_step_image(temp_image/255, folder_to_save_for_paper_figure, name = '4_blob_with_PCA_axes', min_max = [0, 1], draw_line = (X, Y))

        #rotate
        diag = np.sqrt(np.sum(np.asarray(bounding_box_image.shape)**2)).astype(int)
        diag = (diag, diag)
        M = cv2.getRotationMatrix2D(tuple(center), rot_ang, 1)
        minif_rot = cv2.warpAffine(bounding_box_image, M, diag, borderMode=cv2.BORDER_CONSTANT, flags = cv2.INTER_CUBIC)


        crop_distance = int(identification_image_size/2)
        x_range = range(center[0] - crop_distance, center[0] + crop_distance)
        y_range = range(center[1] - crop_distance, center[1] + crop_distance)
        image_for_identification = minif_rot.take(y_range, mode = 'wrap', axis=0).take(x_range, mode = 'wrap', axis=1)
        height, width = image_for_identification.shape

        rot_ang_rad = rot_ang * np.pi / 180
        h_or_t_1 = np.array([np.cos(rot_ang_rad), np.sin(rot_ang_rad)]) * rot_ang_rad
        h_or_t_2 = - h_or_t_1
        if folder_to_save_for_paper_figure:
            save_preprocessing_step_image(image_for_identification/255, folder_to_save_for_paper_figure, name = '5_blob_dilated_rotated', min_max = [0, 1])
        image_for_identification_standarised = ((image_for_identification - np.mean(image_for_identification))/np.std(image_for_identification)).astype('float32')
        if folder_to_save_for_paper_figure:
            save_preprocessing_step_image(image_for_identification_standarised, folder_to_save_for_paper_figure, name = '6_blob_dilated_rotated_normalized', min_max = [np.min(image_for_identification), np.max(image_for_identification)])

        return image_for_identification_standarised, tuple(h_or_t_1.astype('int')), tuple(h_or_t_2.astype('int')), image_for_identification


    """ The following methods are only to be used for the validation and modification
    of trajectories with the pythonvideoannotator after the video is tracked """
    @property
    def contour_full_resolution(self):
        return (self.contour/self.resolution_reduction).astype(np.int32)


    @property
    def bounding_box_full_resolution(self):
        bounding_box_full_resolution = (np.asarray(self.bounding_box_in_frame_coordinates)/self.resolution_reduction).astype(int)
        return tuple(map(tuple, bounding_box_full_resolution))


    @property
    def user_generated_centroids(self):
        if hasattr(self, '_user_generated_centroids'):
            return self._user_generated_centroids
        return None

    @property
    def assigned_centroids(self):
        if hasattr(self, 'interpolated_centroids') and self.interpolated_centroids is not None:
            assert isinstance(self.interpolated_centroids, list)
            return self.interpolated_centroids
        return [self.centroid]

    @property
    def final_centroids(self):
        """
        :return: Return a list of the final centroids.
        """
        if self.user_generated_centroids is not None:
            # Note that sometimes len(user_generated_centroids) > len(assigned_centroids)
            final_centroids = []
            for i, user_generated_centroid in enumerate(self.user_generated_centroids):
                if user_generated_centroid[0] is None:
                    final_centroids.append(self.assigned_centroids[i])
                elif user_generated_centroid[0] > 0:
                    final_centroids.append(user_generated_centroid)
            return final_centroids
        return self.assigned_centroids


    @property
    def final_centroids_full_resolution(self):
        """
        :return: Return a list of the final centroids in the full resolution of the frame
        """
        return [(centroid[0]/self.resolution_reduction,
                 centroid[1]/self.resolution_reduction) for centroid in self.final_centroids]


    def draw(self, image, colors_lst=None, selected_id=None):
        """
        Draw the blob representation in an image
        :param numpy.array image: Image where the blob should be draw.
        :param str selected_id: Identity of the selected blob.
        :param colors_lst: List of colors used to draw the blobs.
        """
        contour = self.contour_full_resolution

        for i, (identity, centroid) in enumerate(zip(self.final_identities, self.final_centroids_full_resolution)):

            pos = int(round(centroid[0], 0)), int(round(centroid[1], 0))

            if colors_lst:
                color = colors_lst[identity] if identity is not None else colors_lst[0]
            else:
                color = (0, 0, 255)

            cv2.polylines(image, np.array([contour]), True, (0, 255, 0), 1)

            cv2.circle(image, pos, 8, (255, 255, 255), -1, lineType=cv2.LINE_AA)
            cv2.circle(image, pos, 6, color, -1, lineType=cv2.LINE_AA)

            if identity is not None:

                if identity == selected_id:
                    cv2.circle(image, pos, 10, (0, 0, 255), 2, lineType=cv2.LINE_AA)

                idroot = ''
                if self.user_generated_identities is not None and (self.user_generated_identities[i] is not None and self.user_generated_identities[i] > 0):
                    idroot = 'u-'
                elif self.user_generated_centroids is not None and (self.user_generated_centroids[i][0] is not None and self.user_generated_centroids[i][0] > 0):
                    idroot = 'u-'
                elif self.identities_corrected_closing_gaps is not None and not self.is_an_individual:
                    idroot = 'c-'


                idstr = idroot + str(identity)
                text_size = cv2.getTextSize(idstr, cv2.FONT_HERSHEY_SIMPLEX, 1.0, thickness=2)
                text_width = text_size[0][0]
                str_pos = pos[0] - text_width // 2, pos[1] - 12

                cv2.putText(image, idstr, str_pos, cv2.FONT_HERSHEY_SIMPLEX, 1.0, (255, 255, 255), thickness=3,
                            lineType=cv2.LINE_AA)
                cv2.putText(image, idstr, str_pos, cv2.FONT_HERSHEY_SIMPLEX, 1.0, color, thickness=2, lineType=cv2.LINE_AA)
            else:
                bounding_box = self.bounding_box_full_resolution
                rect_color   = self.rect_color if hasattr(self, 'rect_color') else (255, 0, 0)
                cv2.rectangle(image, bounding_box[0], bounding_box[1], rect_color, 2)

    def removable_identity(self, identity_to_remove, blobs_in_frame):
        for blob in blobs_in_frame:
            if blob != self:
                if identity_to_remove in blob.final_identities:
                    return True
            else:
                if blob.final_identities.count(identity_to_remove) > 1:
                    return True
        return False

    def update_centroid(self, video, old_centroid, new_centroid):
        """ Updates the coordinates of the centrod

        Parameters
        ----------
        old_centroid : tuple
            len(centroid) must be 2
        new_centroid : tuple
            len(new_centroid) must be 2
        """

        video.is_centroid_updated = True
        old_centroid = (old_centroid[0]*self.resolution_reduction,
                        old_centroid[1]*self.resolution_reduction)
        new_centroid = (new_centroid[0]*self.resolution_reduction,
                        new_centroid[1]*self.resolution_reduction)

        if not (isinstance(old_centroid, tuple) and len(old_centroid) == 2):
            raise Exception("The old_centroid must be a tuple of length 2")
        if not (isinstance(new_centroid, tuple) and len(new_centroid) == 2):
            raise Exception("The new centroid must be a tuple of length 2")

        if self.user_generated_centroids is None:
            self._user_generated_centroids = [(None, None)]*len(self.final_centroids)

        try:
            centroid_index = self.final_centroids.index(old_centroid)
            self.user_generated_centroids[centroid_index] = new_centroid
        except ValueError:
            raise Exception("There is no centroid with the values of old_centroid")

    def remove_centroid(self, video, identity, centroid, blobs_in_frame,
                        apply_resolution_reduction=True):
        """ Remove the centroid and the identity from the blob if it exist.

        Parameters
        ----------
        centroid : tuple
            centroid to be removed in full frame coordinates,
            without the application of the resolution reduction
        """
        if apply_resolution_reduction:
            centroid = (centroid[0]*self.resolution_reduction, centroid[1]*self.resolution_reduction)
        if not (isinstance(centroid, tuple) and len(centroid) == 2):
            raise Exception("The centroid must be a tuple of length 2")

        if not self.removable_identity(identity, blobs_in_frame):
            raise Exception("The centroid cannot be remove beucase it belongs to a unique identity. \
                            Only centroids of duplicated identities can be deleted")

        try:
            centroid_index = self.final_centroids.index(centroid)
        except ValueError:
            raise Exception("Cannot find the centroid in the selected blob")

        if len(self.final_centroids) > 1:
            if self.user_generated_centroids is None:
                self._user_generated_centroids = [(None, None)]*len(self.final_centroids)
            self._user_generated_centroids[centroid_index] = (-1, -1)
            if self.user_generated_identities is None:
                self._user_generated_identities = [None]*len(self.final_identities)
            self._user_generated_identities[centroid_index] = -1
            video.is_centroid_updated = True
        else:
            raise Exception("Cannot remove the centroid of the blob, \
                  because it is the only centroid")

    def add_centroid(self, video, centroid, identity,
                     apply_resolution_reduction=True):
        """ Adds a centroid with a given identity, identity.

        Parameters
        ----------
        centroid : tuple
            centroid to be added in full resolution coordinates. len(centroid) must be 2
        identity : int
            identity of the centroid. identity must be > 0 and <= number_of_animals
        """
        if apply_resolution_reduction:
            centroid = (centroid[0]*self.resolution_reduction,
                        centroid[1]*self.resolution_reduction)
        if not (isinstance(centroid, tuple) and len(centroid) == 2):
            raise Exception("The centroid must be a tuple of length 2")
        if not (isinstance(identity, int) and identity > 0 and identity <= self.number_of_animals):
            raise Exception("The identity must be an integer between 1 and the number of animals in the video")

        if self.user_generated_centroids is None:
            self._user_generated_centroids = [(None, None)]*len(self.final_centroids)
        if self.user_generated_identities is None:
            self._user_generated_identities = [None]*len(self.final_identities)

        self._user_generated_centroids.append(centroid)
        self.user_generated_identities.append(identity)

        video.is_centroid_updated = True

    def update_identity(self, new_id, centroid):
        """ Updates identity. If the blob has multiple identities already assigned
        the old_id to be modified must be specified.

        Parameters
        ----------
        new_id : int
            new value for the identity of the blob
        old_id : int
            old value of the identity of the blob. It must be specified when the
            blob has multiple identities already assigned.
        """
        if not(isinstance(new_id, int) and new_id > 0 and new_id <= self.number_of_animals):
            raise Exception('The new identity must be an integer between 1 and the number of animals in the video')

        if self.user_generated_identities is None:
            self._user_generated_identities = [None]*len(self.final_identities)

        centroid_index = self.final_centroids.index(centroid)

        try:
            centroid_index = self.final_centroids.index(centroid)
            self.user_generated_identities[centroid_index] = new_id
        except ValueError:
            raise Exception('Identity cannot be updated because there is no such centroid in the blob')


    def propagate_identity(self, new_blob_identity, centroid):
        """Propagates the identity specified by new_blob_identity.
        """
        count_past_corrections = 1 #to take into account the modification already done in the current frame
        count_future_corrections = 0

        current = self
        current_centroid = np.asarray(centroid)

        while len(current.next) == 1 and current.next[0].fragment_identifier == self.fragment_identifier:
            if len(current.next[0].final_centroids) > 1:
                next_centroids = np.asarray(current.next[0].centroids)
                index_centroid = np.argmin(np.sqrt(np.sum((current_centroid-next_centroids)**2)))
                next_centroid = current.next[0].final_centroids[index_centroid]
            else:
                next_centroid = current.next[0].final_centroids[0]
            current.next[0].update_identity(new_blob_identity, next_centroid)
            current = current.next[0]
            current_centroid = np.asarray(next_centroid)
            count_future_corrections += 1

        current = self

        while len(current.previous) == 1 and current.previous[0].fragment_identifier == self.fragment_identifier:
            if len(current.previous[0].final_centroids) > 1:
                previous_centroids = np.asarray(current.previous[0].centroids)
                index_centroid = np.argmin(np.sqrt(np.sum((current_centroid-previous_centroids)**2)))
                previous_centroid = current.previous[0].final_centroids[index_centroid]
            else:
                previous_centroid = current.previous[0].final_centroids[0]
            current.previous[0].update_identity(new_blob_identity, previous_centroid)
            current = current.previous[0]
            current_centroid = np.asarray(previous_centroid)
            count_past_corrections += 1


def remove_background_pixels(height, width, bounding_box_image, pixels,
                             bounding_box_in_frame_coordinates,
                             folder_to_save_for_paper_figure):
    """Removes the background pixels substiuting them with a homogeneous black
    background.

    Parameters
    ----------
    height : int
        Frame height
    width : int
        Frame width
    bounding_box_image : ndarray
        Images cropped from the frame by considering the bounding box associated to a blob
    pixels : list
        List of pixels associated to a blob
    bounding_box_in_frame_coordinates : list
        [(x, y), (x + bounding_box_width, y + bounding_box_height)]
    identification_image_size : tuple
        shape of the identification image
    folder_to_save_for_paper_figure : str
        folder to save the images for identification

    Returns
    -------
    ndarray
        Image with black background pixels

    """
    pxs = np.array(np.unravel_index(pixels, (height, width))).T
    pxs = np.array([pxs[:, 0] - bounding_box_in_frame_coordinates[0][1], pxs[:, 1] - bounding_box_in_frame_coordinates[0][0]])
    temp_image = np.zeros_like(bounding_box_image).astype('uint8')
    temp_image[pxs[0,:], pxs[1,:]] = 255
    if folder_to_save_for_paper_figure:
        new_thresholded_image = temp_image.copy()
        new_bounding_box_image = bounding_box_image.copy()
        new_bounding_box_image = cv2.cvtColor(new_bounding_box_image, cv2.COLOR_GRAY2RGB)
        _, contours, hierarchy = cv2.findContours(new_thresholded_image, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        cv2.drawContours(new_bounding_box_image, contours, -1, (255,0,0), 1)
        save_preprocessing_step_image(new_bounding_box_image/255, folder_to_save_for_paper_figure, name = '1_blob_contour',  min_max = [0, 1])
    temp_image = cv2.dilate(temp_image, np.ones((3,3)).astype('uint8'), iterations = 1)
    if folder_to_save_for_paper_figure:
        new_thresholded_image = temp_image.copy()
        new_bounding_box_image = bounding_box_image.copy()
        new_bounding_box_image = cv2.cvtColor(new_bounding_box_image, cv2.COLOR_GRAY2RGB)
        _, contours, hierarchy = cv2.findContours(new_thresholded_image, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        cv2.drawContours(new_bounding_box_image, contours, -1, (255,0,0), 1)
        save_preprocessing_step_image(new_bounding_box_image/255, folder_to_save_for_paper_figure, name = '2_blob_bw_dilated',  min_max = [0, 1])
    rows, columns = np.where(temp_image == 255)
    dilated_pixels = np.array([rows, columns])
    temp_image[dilated_pixels[0,:], dilated_pixels[1,:]] = bounding_box_image[dilated_pixels[0,:], dilated_pixels[1,:]]
    if folder_to_save_for_paper_figure:
        save_preprocessing_step_image(temp_image/255, folder_to_save_for_paper_figure, name = '3_blob_contour_dilated',  min_max = [0, 1])
    return temp_image

def full2miniframe(point, boundingBox):
    """Maps a point in the fullframe to the coordinate system defined by the image
    generated by considering the bounding box of the blob.
    Here it is use for centroids

    Parameters
    ----------
    point : tuple
        (x, y)
    boundingBox : list
        [(x, y), (x + bounding_box_width, y + bounding_box_height)]

    Returns
    -------
    tuple
        :math:`(x^\prime, y^\prime)`

    """
    return tuple(np.asarray(point) - np.asarray([boundingBox[0][0],boundingBox[0][1]]))

def save_preprocessing_step_image(image, save_folder, name = None, min_max = None, draw_line = None):
    from matplotlib import pyplot as plt
    import seaborn as sns
    fig, ax = plt.subplots(1,1)
    ax.imshow(image, cmap = 'gray', vmin = min_max[0], vmax = min_max[1])
    if draw_line is not None:
        plt.plot(draw_line[0], draw_line[1], 'b-')
        ax.set_xlim([0, image.shape[1]])
        ax.set_ylim([0, image.shape[0]])
        ax.invert_yaxis()

    ax.set_xticks([])
    ax.set_yticks([])

    sns.despine(left = True, right = True, top = True, bottom = True)
    fig.savefig(os.path.join(save_folder,'%s.pdf' %name), transparent=True, bbox_inches='tight')



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

import numpy as np
from tqdm import tqdm

from idtrackerai.list_of_blobs import ListOfBlobs

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.get_trajectories")

"""
Usage: get_trajectories.py

Contains tools to process tracked ListOfBlobs
and output trajectories as numpy files with dimensions:

    [Frame number x Individual number  x  coordinate (x,y)]

When a certain individual was not identified in the frame
a NaN appears instead of the coordinates
"""

def assign_point_to_identity(centroid, identity, frame_number, centroid_trajectories):
    """Populate the matrix of individual trajectories with the centroid of a
    selected Blob object (see :class:`~blob.Blob`)

    Parameters
    ----------
    centroid : tuple
        (x, y)
    identity : int
        Identity to be associated with the centroid
    frame_number : int
        Frame number in the tracked video
    centroid_trajectories : ndarray
        array of shape [number of frame in video x number of animals  x  2]

    Returns
    -------
    ndarray
        centroid_trajectories

    """
    if identity is not None and identity != 0:
        centroid_trajectories[frame_number, identity - 1, :] = centroid
    return centroid_trajectories


def assign_P2_to_identity(P2_vector, identity, frame_number, id_probabilities):
    """Populate the matrix of P2 trajectories with the argmax of the P2_vector
    of a selected Blob object (see :class:`~blob.Blob`)

    Parameters
    ----------
    P2_vector : array
        Array with P2 values for a Blob
    identity : int
        Identity to be associated with the centroid
    frame_number : int
        Frame number in the tracked video
    centroid_trajectories : ndarray
        array of shape [number of frame in video x number of animals  x  2]

    Returns
    -------
    ndarray
        centroid_trajectories

    """
    if identity is not None and identity != 0:
        id_probabilities[frame_number, identity - 1, :] = np.max(P2_vector)
    return id_probabilities


def produce_trajectories(blobs_in_video, number_of_frames, number_of_animals):
    """Produce trajectories array from ListOfBlobs

    Parameters
    ----------
    blobs_in_video : <ListOfBlobs object>
        See :class:`list_of_blobs.ListOfBlobs`
    number_of_frames : int
        Total number of frames in video
    number_of_animals : int
        Number of animals to be tracked

    Returns
    -------
    dict
        Dictionary with np.array as values (trajectories organised by identity)

    """
    centroid_trajectories = np.ones((number_of_frames, number_of_animals, 2)) * np.NaN
    id_probabilities = np.ones((number_of_frames, number_of_animals, 1)) * np.NaN

    for frame_number, blobs_in_frame in enumerate(tqdm(blobs_in_video)):

        for blob in blobs_in_frame:
            for identity, centroid in zip(blob.final_identities, blob.final_centroids_full_resolution):
                centroid_trajectories = assign_point_to_identity(centroid,
                                                                identity,
                                                                blob.frame_number,
                                                                centroid_trajectories)
            if blob.is_an_individual and len(blob.final_identities) == 1 and hasattr(blob, '_P2_vector') and blob._P2_vector is not None:
                id_probabilities = assign_P2_to_identity(blob._P2_vector,
                                                         blob.final_identities[0],
                                                         blob.frame_number,
                                                         id_probabilities)

    return centroid_trajectories, id_probabilities


def produce_trajectories_wo_identities(blobs_in_video, number_of_frames, number_of_animals):
    centroid_trajectories = np.ones((number_of_frames, number_of_animals, 2))*np.nan
    identifiers_prev = np.arange(number_of_animals).astype(np.float32)
    for frame_number, blobs_in_frame in enumerate(tqdm(blobs_in_video, "creating trajectories")):
        if frame_number != len(blobs_in_video)-1:
            identifiers_next = [b.fragment_identifier for b in blobs_in_video[frame_number+1]]
        else:
            identifiers_next = [b.fragment_identifier for b in blobs_in_video[frame_number]]
        for blob_number, blob in enumerate(blobs_in_frame):
            if blob.is_an_individual:
                if blob.fragment_identifier in identifiers_prev:
                    column = np.where(identifiers_prev == blob.fragment_identifier)[0][0]
                else:
                    column = np.where(np.isnan(identifiers_prev))[0][0]
                    identifiers_prev[column] = blob.fragment_identifier

                blob._identity = int(column+1)
                centroid_trajectories[frame_number, column, :] = blob.final_centroids_full_resolution[0] # blobs that are individual only have one centroid

                if blob.fragment_identifier not in identifiers_next:
                    identifiers_prev[column] = np.nan
    return centroid_trajectories, None


def produce_output_dict(blobs_in_video, video):
    """Outputs the dictionary with keys: trajectories, git_commit, video_path,
    frames_per_second

    Parameters
    ----------
    blobs_in_video : list
        List of all blob objects (see :class:`~blob.Blobs`) generated by
        considering all the blobs segmented from the video
    video : <Video object>
        See :class:`~video.Video`

    Returns
    -------
    dict
        Output dictionary containing trajectories as values

    """
    if not video.track_wo_identities:
        centroid_trajectories, id_probabilities = \
            produce_trajectories(blobs_in_video, video.number_of_frames,
                                 video.number_of_animals)
    else:
        video._number_of_animals = np.max([len(bf) for bf in blobs_in_video])
        centroid_trajectories, id_probabilities = \
            produce_trajectories_wo_identities(blobs_in_video,
                                                video.number_of_frames,
                                                video.number_of_animals)
    output_dict = {'trajectories': centroid_trajectories,
                   'id_probabilities': id_probabilities,
                   'git_commit': video.git_commit,
                   'video_path': video.video_path,
                   'frames_per_second': video.frames_per_second,
                   'body_length': video.median_body_length_full_resolution}
    return output_dict



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import numpy as np
from confapp import conf

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.get_data")

np.random.seed(0)


class DataSet(object):
    """Contains the `images` and `labels` to be used for training a particular
    model

    Attributes
    ----------

    images : ndarray
        Array of shape [num_images, height, width, channels] containing
        the images of the dataset
    num_images : int
        Number of images in the dataset
    labels : ndarray
        Array of shape [num_images, number_of_classes] containing the
        labels corresponding to the images in the dataset
    number_of_animals : int
        Number of classes in the dataset

    """
    def __init__(self, number_of_animals = None, images = None, labels = None):
        self.images = images
        self._num_images = len(self.images)
        self.labels = labels
        self.number_of_animals = number_of_animals
        #check the number of images and labels are the same. If it true set the num_images
        self.consistency_check()

    def consistency_check(self):
        """Checks that the length of :attr:`images` and :attr:`labels` is the same
        """
        if self.labels is not None:
            assert len(self.images) == len(self.labels)

    def convert_labels_to_one_hot(self):
        """Converts labels from dense format to one hot format
        See Also
        --------
        :func:`~get_data.shuffle_images_and_labels`
        """
        self.labels = dense_to_one_hot(self.labels, n_classes=self.number_of_animals)


def duplicate_PCA_images(training_images, training_labels):
    """Creates a copy of every image in `training_images` by rotating 180 degrees

    Parameters
    ----------
    training_images : ndarray
        Array of shape [number of images, height, width, channels] containing
        the images to be rotated
    training_labels : ndarray
        Array of shape [number of images, 1] containing the labels corresponding
        to the `training_images`

    Returns
    -------
    training_images : ndarray
        Array of shape [2*number of images, height, width, channels] containing
        the original images and the images rotated
    training_labels : ndarray
        Array of shape [2*number of images, 1] containing the labels corresponding
        to the original images and the images rotated
    """
    augmented_images = [np.rot90(image, 2) for image in training_images]
    training_images = np.concatenate([training_images, augmented_images], axis = 0)
    training_labels = np.concatenate([training_labels, training_labels], axis = 0)
    return training_images, training_labels

def split_data_train_and_validation(number_of_animals, images, labels, validation_proportion = conf.VALIDATION_PROPORTION):
    """Splits a set of `images` and `labels` into training and validation sets

    Parameters
    ----------
    number_of_animals : int
        Number of classes in the set of images
    images : list
        List of images (arrays of shape [height, width])
    labels : list
        List of integers from 0 to `number_of_animals` - 1
    validation_proportion : float
        The proportion of images that will be used to create the validation set.


    Returns
    -------
    training_dataset : <DataSet object>
        Object containing the images and labels for training
    validation_dataset : <DataSet object>
        Object containing the images and labels for validation

    See Also
    --------
    :class:`get_data.DataSet`
    :func:`get_data.duplicate_PCA_images`
    :func:`get_data.shuffle_images_and_labels`
    """
    # Init variables
    train_images = []
    train_labels = []
    validation_images = []
    validation_labels = []
    images = np.expand_dims(np.asarray(images), axis = 3)
    labels = np.expand_dims(np.asarray(labels), axis = 1)
    images, labels = shuffle_images_and_labels(images, labels)
    for i in np.unique(labels):
        # Get images of this individual
        this_indiv_images = images[np.where(labels == i)[0]]
        this_indiv_labels = labels[np.where(labels == i)[0]]
        # Compute number of images for training and validation
        num_images = len(this_indiv_labels)
        num_images_validation = np.ceil(validation_proportion*num_images).astype(int)
        num_images_training = num_images - num_images_validation
        # Get train, validation and test, images and labels
        train_images.append(this_indiv_images[:num_images_training])
        train_labels.append(this_indiv_labels[:num_images_training])
        validation_images.append(this_indiv_images[num_images_training:])
        validation_labels.append(this_indiv_labels[num_images_training:])

    train_images = np.vstack(train_images)
    train_labels = np.vstack(train_labels)
    train_images, train_labels = duplicate_PCA_images(train_images, train_labels)
    train_images, train_labels = shuffle_images_and_labels(train_images, train_labels)
    validation_images = np.vstack(validation_images)
    validation_labels = np.vstack(validation_labels)
    return DataSet(number_of_animals, train_images, train_labels), DataSet(number_of_animals, validation_images, validation_labels)

def shuffle_images_and_labels(images, labels):
    """Shuffles images and labels with a random
    permutation, according to the number of examples"""
    np.random.seed(0)
    perm = np.random.permutation(len(labels))
    images = images[perm]
    labels = labels[perm]
    return images, labels

def dense_to_one_hot(labels, n_classes=2):
    """Convert class labels from scalars to one-hot vectors."""
    labels = np.array(labels)
    n_labels = labels.shape[0]
    index_offset = np.arange(n_labels) * n_classes
    labels_one_hot = np.zeros((n_labels, n_classes), dtype=np.float32)
    labels_one_hot.flat[index_offset + labels.ravel()] = 1
    return labels_one_hot



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import tensorflow as tf
import os
import numpy as np

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.crossings_detector_model")

class ConvNetwork_crossings(object):
    def __init__(self, network_params):

        self.params = network_params
        self.architecture = network_params.architecture
        self.learning_rate = network_params.learning_rate
        self.image_size = network_params.image_size
        self.number_of_classes = network_params.number_of_classes
        logger.info("Building graph....")
        self._build_graph()
        self.ops_list = [self.loss, self.accuracy, self.individual_accuracy]
        self.saver = tf.train.Saver()
        self.sesh = tf.Session()
        self.sesh.run(tf.global_variables_initializer())

    def _build_graph(self):
        self.X = tf.placeholder(tf.float32, shape = [None, self.image_size[0], self.image_size[1], self.image_size[2]])
        self.Y_logits = self.architecture(self.X, self.number_of_classes, self.image_size[0], self.image_size[1], self.image_size[2])
        self.Y_target = tf.placeholder(tf.float32, shape = [None, 2])
        self.loss_weights_pl = tf.placeholder(tf.float32, [None], name = 'loss_weights')
        self.loss = self.weighted_loss()
        self.train_step = tf.train.AdamOptimizer(self.learning_rate).minimize(self.loss)
        # train_step = tf.train.GradientDescentOptimizer(self.learning_rate).minimize(loss)
        self.accuracy = self.compute_accuracy(self.Y_target, self.Y_logits)
        self.individual_accuracy = self.compute_individual_accuracy(self.Y_target, self.Y_logits, self.number_of_classes)
        self.softmax_probs = tf.nn.softmax(self.Y_logits)
        self.predictions = tf.cast(tf.argmax(self.softmax_probs,1),tf.float32)

    def weighted_loss(self):
        cross_entropy = tf.reduce_mean(
            tf.losses.softmax_cross_entropy(self.Y_target, self.Y_logits, self.loss_weights_pl), name = 'CrossEntropyMean')
        return cross_entropy

    def compute_loss_weights(self, training_labels):
        self.weights = 1. - np.sum(training_labels, axis=0) / len(training_labels)

    @staticmethod
    def compute_accuracy(labels, logits):
        # We add 1 to the labels and predictions to avoid having a 0 label
        labels = tf.cast(tf.add(tf.where(tf.equal(labels,1))[:,1],1),tf.float32)
        predictions = tf.cast(tf.add(tf.argmax(logits,1),1),tf.float32)
        # acc = tf.metrics.accuracy(labels, predictions)
        correct_prediction = tf.equal(predictions, labels, name='correctPrediction')
        acc = tf.reduce_mean(tf.cast(correct_prediction, 'float'), name='overallAccuracy')
        return acc

    @staticmethod
    def compute_individual_accuracy(labels,logits,classes):
        # We add 1 to the labels and predictions to avoid having a 0 label
        labels = tf.cast(tf.add(tf.where(tf.equal(labels,1))[:,1],1),tf.float32)
        predictions = tf.cast(tf.add(tf.argmax(logits,1),1),tf.float32)
        labelsRep = tf.reshape(tf.tile(labels, [classes]), [classes,tf.shape(labels)[0]])

        correct = tf.cast(tf.equal(labels,predictions),tf.float32)
        indivCorrect = tf.multiply(predictions,correct)

        indivRep = tf.cast(tf.transpose(tf.reshape(tf.tile(tf.range(1,classes+1), [tf.shape(labels)[0]]), [tf.shape(labels)[0],classes])),tf.float32)
        indivCorrectRep = tf.reshape(tf.tile(indivCorrect, [classes]), [classes,tf.shape(labels)[0]])
        correctPerIndiv = tf.cast(tf.equal(indivRep,indivCorrectRep),tf.float32)

        countCorrect = tf.reduce_sum(correctPerIndiv,1)
        numImagesPerIndiv = tf.reduce_sum(tf.cast(tf.equal(labelsRep,indivRep),tf.float32),1)

        indivAcc = tf.div(countCorrect,numImagesPerIndiv)

        return indivAcc

    def compute_batch_weights(self, batch_labels):
        batch_weights = np.sum(self.weights*batch_labels,axis=1)
        return batch_weights

    def train(self,batch):
        (batch_images, batch_labels) = batch
        batch_weights = self.compute_batch_weights(batch_labels)
        feed_dict={self.X: batch_images, self.Y_target: batch_labels, self.loss_weights_pl: batch_weights}
        self.sesh.run(self.train_step, feed_dict = feed_dict)
        out_list = self.sesh.run(self.ops_list, feed_dict = feed_dict)
        return out_list, feed_dict

    def validate(self,batch):
        (batch_images, batch_labels) = batch
        batch_weights = self.compute_batch_weights(batch_labels)
        feed_dict={self.X: batch_images, self.Y_target: batch_labels, self.loss_weights_pl: batch_weights}
        self.sesh.run(self.train_step, feed_dict = feed_dict)
        out_list = self.sesh.run(self.ops_list, feed_dict = feed_dict)
        return out_list, feed_dict

    def prediction(self,batch_images):
        return self.sesh.run(self.predictions, feed_dict={self.X: batch_images})

    @property
    def is_knowledge_transfer(self):
        return self.params._knowledge_transfer_folder is not None

    @property
    def is_restoring(self):
        return self.params._restore_folder is not None

    def restore(self):
        self.sesh.run(tf.global_variables_initializer())
        try:
            ckpt = tf.train.get_checkpoint_state(self.params._restore_folder)
            logger.info("restoring crossings detector model from %s" %ckpt.model_checkpoint_path)
            self.saver.restore(self.sesh, ckpt.model_checkpoint_path) # restore convolutional variables
        except:
            logger.info('\nWarning: no checkpoints found in the folder %s' %self.params._restore_folder)

    def save(self, global_step):
        logger.info("(in save) self.params._save_folder %s" %self.params._save_folder)
        save_path = self.saver.save(self.sesh, os.path.join(self.params._save_folder,'.ckpt'), global_step=global_step)
        logger.info("Model saved in file: %s" %save_path)
